import os
import unittest
from datetime import datetime, timedelta
from datetime import timezone as dt_timezone
from unittest import mock

from django.conf import settings
from django.template import Context, Template, TemplateSyntaxError
from django.test import SimpleTestCase, TestCase
from django.test.utils import override_settings
from django.utils import timezone
from freezegun import freeze_time

from wagtail.admin.staticfiles import VERSION_HASH, versioned_static
from wagtail.admin.templatetags.wagtailadmin_tags import (
    absolute_static,
    avatar_url,
    i18n_enabled,
    locale_label_from_id,
    timesince_last_update,
    timesince_simple,
)
from wagtail.coreutils import get_dummy_request
from wagtail.images.tests.utils import get_test_image_file
from wagtail.models import Locale, Page
from wagtail.test.utils import WagtailTestUtils
from wagtail.test.utils.template_tests import AdminTemplateTestUtils
from wagtail.users.models import UserProfile


class TestAvatarUrlInterceptTemplateTag(WagtailTestUtils, TestCase):
    def setUp(self):
        self.test_user = self.create_user(
            username="testuser",
            email="testuser@email.com",
            password="password",
        )

    def test_get_avatar_url_undefined(self):
        url = avatar_url(self.test_user)
        self.assertIn("www.gravatar.com", url)

    @mock.patch.dict(os.environ, {"AVATAR_INTERCEPT": "True"}, clear=True)
    def test_get_avatar_url_registered(self):
        url = avatar_url(self.test_user)
        self.assertEqual(url, "/some/avatar/fred.png")


class TestAvatarTemplateTag(WagtailTestUtils, TestCase):
    def setUp(self):
        # Create a user
        self.test_user = self.create_user(
            username="testuser",
            email="testuser@email.com",
            password="password",
        )

    def test_use_gravatar_by_default(self):
        url = avatar_url(self.test_user)
        self.assertIn("www.gravatar.com", url)

    def test_skip_gravatar_if_no_email(self):
        self.test_user.email = ""
        url = avatar_url(self.test_user)
        self.assertIn("default-user-avatar", url)

    @override_settings(WAGTAIL_GRAVATAR_PROVIDER_URL="https://robohash.org")
    def test_custom_gravatar_provider(self):
        url = avatar_url(self.test_user)
        self.assertIn("robohash.org", url)

    @override_settings(WAGTAIL_GRAVATAR_PROVIDER_URL=None)
    def test_disable_gravatar(self):
        url = avatar_url(self.test_user)
        self.assertIn("default-user-avatar", url)

    def test_uploaded_avatar(self):
        user_profile = UserProfile.get_for_user(self.test_user)
        user_profile.avatar = get_test_image_file(filename="custom-avatar.png")
        user_profile.save()

        url = avatar_url(self.test_user)
        self.assertIn("custom-avatar", url)


class TestAbsoluteStaticTemplateTag(SimpleTestCase):
    @override_settings(STATIC_URL="/static/")
    def test_local_absolute_static(self):
        url = absolute_static("wagtailadmin/images/email-header.jpg")
        expected = (
            rf"^{settings.WAGTAILADMIN_BASE_URL}/static/wagtailadmin/images/"
            r"email-header.jpg\?v=(\w{8})$"
        )
        self.assertRegex(url, expected)

    @override_settings(
        STATIC_URL="/static/", WAGTAILADMIN_BASE_URL="http://localhost:8000"
    )
    def test_local_absolute_static_baseurl(self):
        url = absolute_static("wagtailadmin/images/email-header.jpg")
        expected = (
            r"^http://localhost:8000/static/wagtailadmin/images/"
            r"email-header.jpg\?v=(\w{8})$"
        )
        self.assertRegex(url, expected)

    @override_settings(
        STATIC_URL="https://s3.amazonaws.com/somebucket/static/",
        WAGTAILADMIN_BASE_URL="http://localhost:8000",
    )
    def test_remote_absolute_static(self):
        url = absolute_static("wagtailadmin/images/email-header.jpg")
        expected = (
            r"https://s3.amazonaws.com/somebucket/static/wagtailadmin/images/"
            r"email-header.jpg\?v=(\w{8})$"
        )
        self.assertRegex(url, expected)


class TestVersionedStatic(SimpleTestCase):
    def test_version_hash(self):
        self.assertEqual(len(VERSION_HASH), 8)

    def test_versioned_static(self):
        result = versioned_static("wagtailadmin/js/core.js")
        self.assertRegex(result, r"^/static/wagtailadmin/js/core.js\?v=(\w{8})$")

    @mock.patch("wagtail.admin.staticfiles.static")
    def test_versioned_static_version_string(self, mock_static):
        mock_static.return_value = "/static/wagtailadmin/js/core.js?v=123"
        result = versioned_static("wagtailadmin/js/core.js")
        self.assertEqual(result, "/static/wagtailadmin/js/core.js?v=123")
        mock_static.assert_called_once_with("wagtailadmin/js/core.js")

    def test_versioned_static_absolute_path(self):
        result = versioned_static("/static/wagtailadmin/js/core.js")
        self.assertEqual(result, "/static/wagtailadmin/js/core.js")

    def test_versioned_static_url(self):
        result = versioned_static("http://example.org/static/wagtailadmin/js/core.js")
        self.assertEqual(result, "http://example.org/static/wagtailadmin/js/core.js")


class TestTimesinceTags(SimpleTestCase):
    # timezone matches TIME_ZONE = "Asia/Tokyo" in tests/settings.py
    @freeze_time("2020-07-01 12:00:00+09:00")
    def test_timesince_simple(self):
        now = timezone.make_aware(
            datetime(2020, 7, 1, 12, 0, 0)
        )  # aware date in Asia/Tokyo
        ts = timesince_simple(now)
        self.assertEqual(ts, "just now")

        now = timezone.make_aware(
            datetime(2020, 7, 1, 3, 0, 0), timezone=dt_timezone.utc
        )  # aware date in UTC
        ts = timesince_simple(now)
        self.assertEqual(ts, "just now")

        seventy_minutes_ago = timezone.make_aware(datetime(2020, 7, 1, 10, 50, 0))
        ts = timesince_simple(seventy_minutes_ago)
        self.assertEqual(ts, "1\xa0hour ago")

        two_weeks_ago = timezone.make_aware(datetime(2020, 6, 17, 10, 50, 0))
        ts = timesince_simple(two_weeks_ago)
        self.assertEqual(ts, "2\xa0weeks ago")

    @unittest.skipIf(not settings.USE_TZ, "Test assumes timezone support is active")
    @freeze_time("2020-07-01 06:00:00+09:00")
    def test_timesince_last_update_today_shows_time(self):
        one_hour_ago = timezone.make_aware(
            datetime(2020, 7, 1, 5, 0, 0)
        )  # aware date in Asia/Tokyo
        timesince = timesince_last_update(one_hour_ago)
        self.assertEqual(timesince, "05:00")

        # Check prefix output
        timesince = timesince_last_update(one_hour_ago, show_time_prefix=True)
        self.assertEqual(timesince, "at 05:00")

        # Check user output
        timesince = timesince_last_update(one_hour_ago, user_display_name="Gary")
        self.assertEqual(timesince, "05:00 by Gary")

        # Check user and prefix output
        timesince = timesince_last_update(
            one_hour_ago, show_time_prefix=True, user_display_name="Gary"
        )
        self.assertEqual(timesince, "at 05:00 by Gary")

        one_hour_ago = timezone.make_aware(
            datetime(2020, 6, 30, 20, 0, 0), timezone=dt_timezone.utc
        )  # aware date in UTC
        timesince = timesince_last_update(one_hour_ago)
        self.assertEqual(timesince, "05:00")

    @unittest.skipIf(settings.USE_TZ, "Test assumes timezone support is disabled")
    @freeze_time("2020-07-01 06:00:00")
    def test_timesince_last_update_today_shows_time_without_tz(self):
        one_hour_ago = datetime(2020, 7, 1, 5, 0, 0)
        timesince = timesince_last_update(one_hour_ago)
        self.assertEqual(timesince, "05:00")

        # Check prefix output
        timesince = timesince_last_update(one_hour_ago, show_time_prefix=True)
        self.assertEqual(timesince, "at 05:00")

        # Check user output
        timesince = timesince_last_update(one_hour_ago, user_display_name="Gary")
        self.assertEqual(timesince, "05:00 by Gary")

        # Check user and prefix output
        timesince = timesince_last_update(
            one_hour_ago, show_time_prefix=True, user_display_name="Gary"
        )
        self.assertEqual(timesince, "at 05:00 by Gary")

    @unittest.skipIf(not settings.USE_TZ, "Test assumes timezone support is active")
    @freeze_time("2020-07-01 06:00:00+09:00")
    def test_timesince_last_update_before_midnight_shows_timeago(self):
        """
        If the last update was yesterday in local time, we show "x hours ago" even if it was less
        than 24 hours ago (and even if it matches today's date in UTC)
        """
        eight_hours_ago = timezone.make_aware(
            datetime(2020, 6, 30, 21, 50, 0)
        )  # aware date in Asia/Tokyo
        timesince = timesince_last_update(eight_hours_ago)
        self.assertEqual(timesince, "8\xa0hours ago")

    @unittest.skipIf(settings.USE_TZ, "Test assumes timezone support is disabled")
    @freeze_time("2020-07-01 06:00:00")
    def test_timesince_last_update_before_midnight_shows_timeago_without_tz(self):
        """
        If the last update was yesterday in local time, we show "x hours ago" even if it was less
        than 24 hours ago
        """
        eight_hours_ago = datetime(2020, 6, 30, 21, 50, 0)
        timesince = timesince_last_update(eight_hours_ago)
        self.assertEqual(timesince, "8\xa0hours ago")

    @unittest.skipIf(not settings.USE_TZ, "Test assumes timezone support is active")
    @freeze_time("2020-07-01 12:00:00+09:00")
    def test_timesince_last_update_before_today_shows_timeago(self):
        dt = timezone.make_aware(datetime(2020, 6, 22, 12, 0, 0))

        # 1) use_shorthand=False

        timesince = timesince_last_update(dt, use_shorthand=False)
        self.assertEqual(timesince, "1\xa0week, 2\xa0days ago")
        # The prefix is not used, if the date is older than the current day.
        self.assertEqual(
            timesince_last_update(dt, use_shorthand=False, show_time_prefix=True),
            timesince,
        )

        # Check user output
        timesince = timesince_last_update(
            dt, use_shorthand=False, user_display_name="Gary"
        )
        self.assertEqual(timesince, "1\xa0week, 2\xa0days ago by Gary")
        self.assertEqual(
            timesince_last_update(
                dt, use_shorthand=False, user_display_name="Gary", show_time_prefix=True
            ),
            timesince,
        )

        # 2) use_shorthand=True

        timesince = timesince_last_update(dt)
        self.assertEqual(timesince, "1\xa0week ago")
        self.assertEqual(timesince_last_update(dt, show_time_prefix=True), timesince)

        timesince = timesince_last_update(dt, user_display_name="Gary")
        self.assertEqual(timesince, "1\xa0week ago by Gary")
        self.assertEqual(
            timesince_last_update(dt, user_display_name="Gary", show_time_prefix=True),
            timesince,
        )

    @unittest.skipIf(settings.USE_TZ, "Test assumes timezone support is disabled")
    @freeze_time("2020-07-01 12:00:00")
    def test_timesince_last_update_before_today_shows_timeago_without_tz(self):
        dt = timezone.make_aware(datetime(2020, 6, 22, 12, 0, 0))

        # 1) use_shorthand=False

        timesince = timesince_last_update(dt, use_shorthand=False)
        self.assertEqual(timesince, "1\xa0week, 2\xa0days ago")
        # The prefix is not used, if the date is older than the current day.
        self.assertEqual(
            timesince_last_update(dt, use_shorthand=False, show_time_prefix=True),
            timesince,
        )

        # Check user output
        timesince = timesince_last_update(
            dt, use_shorthand=False, user_display_name="Gary"
        )
        self.assertEqual(timesince, "1\xa0week, 2\xa0days ago by Gary")
        self.assertEqual(
            timesince_last_update(
                dt, use_shorthand=False, user_display_name="Gary", show_time_prefix=True
            ),
            timesince,
        )

        # 2) use_shorthand=True

        timesince = timesince_last_update(dt)
        self.assertEqual(timesince, "1\xa0week ago")
        self.assertEqual(timesince_last_update(dt, show_time_prefix=True), timesince)

        timesince = timesince_last_update(dt, user_display_name="Gary")
        self.assertEqual(timesince, "1\xa0week ago by Gary")
        self.assertEqual(
            timesince_last_update(dt, user_display_name="Gary", show_time_prefix=True),
            timesince,
        )

    @override_settings(USE_TZ=False)
    @freeze_time("2020-07-01 12:00:00")
    def test_human_readable_date(self):
        now = timezone.now()
        template = """
            {% load wagtailadmin_tags %}
            {% human_readable_date date %}
        """

        html = Template(template).render(Context({"date": now}))
        self.assertIn("Just now", html)

        html = Template(template).render(
            Context({"date": now - timedelta(hours=1, minutes=10)})
        )
        self.assertIn("1\xa0hour ago", html)
        self.assertIn('data-w-tooltip-placement-value="top"', html)
        self.assertIn('data-w-tooltip-content-value="July 1, 2020, 10:50 a.m."', html)

    @override_settings(USE_TZ=False)
    @freeze_time("2020-07-01 12:00:00")
    def test_human_readable_date_with_date_object(self):
        today = timezone.now().date()
        template = """
            {% load wagtailadmin_tags %}
            {% human_readable_date date %}
        """

        html = Template(template).render(Context({"date": today}))
        self.assertIn("12\xa0hours ago", html)

        html = Template(template).render(
            Context({"date": today - timedelta(days=1, hours=1)})
        )
        self.assertIn("1\xa0day ago", html)
        self.assertIn('data-w-tooltip-placement-value="top"', html)
        self.assertIn('data-w-tooltip-content-value="June 30, 2020"', html)

    @freeze_time("2020-07-01 12:00:00")
    def test_human_readable_date_with_args(self):
        now = timezone.now()
        template = """
            {% load wagtailadmin_tags %}
            {% human_readable_date date "The clock ticked" "bottom" %}
        """

        html = Template(template).render(Context({"date": now}))
        self.assertIn(
            '<span class="w-human-readable-date__description">The clock ticked</span>',
            html,
        )
        self.assertIn('data-w-tooltip-placement-value="bottom"', html)


@override_settings(
    WAGTAIL_CONTENT_LANGUAGES=[
        ("en", "English"),
        ("fr", "French"),
        ("ro", "Romanian"),
        ("ru", "Russian"),
    ],
    CACHES={"default": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"}},
)
class TestInternationalisationTags(TestCase):
    def setUp(self):
        self.locale_ids = []
        for language_code in ["en", "fr", "ro", "ru"]:
            locale, _ = Locale.objects.get_or_create(language_code=language_code)
            self.locale_ids.append(locale.pk)

    def test_i18n_enabled(self):
        with override_settings(WAGTAIL_I18N_ENABLED=False):
            self.assertFalse(i18n_enabled())

        with override_settings(WAGTAIL_I18N_ENABLED=True):
            self.assertTrue(i18n_enabled())

    def test_locale_label_from_id(self):
        with self.assertNumQueries(1):
            self.assertEqual(locale_label_from_id(self.locale_ids[0]), "English")

        with self.assertNumQueries(0):
            self.assertEqual(locale_label_from_id(self.locale_ids[1]), "French")

        # check with an invalid id
        with self.assertNumQueries(0):
            self.assertIsNone(locale_label_from_id(self.locale_ids[-1] + 100), None)


class ComponentTest(SimpleTestCase):
    def test_render_block_component(self):
        template = """
            {% load wagtailadmin_tags %}
            {% help_block status="info" %}Proceed with caution{% endhelp_block %}
        """

        expected = """
            <div class="help-block help-info">
                <svg aria-hidden="true" class="icon icon icon-help"><use href="#icon-help"></svg>
                Proceed with caution
            </div>
        """

        self.assertHTMLEqual(expected, Template(template).render(Context()))

    def test_render_nested(self):
        template = """
            {% load wagtailadmin_tags %}
            {% help_block status="warning" %}
                {% help_block status="info" %}Proceed with caution{% endhelp_block %}
            {% endhelp_block %}
        """

        expected = """
            <div class="help-block help-warning">
                <svg aria-hidden="true" class="icon icon icon-warning"><use href="#icon-warning"></svg>
                <div class="help-block help-info">
                    <svg aria-hidden="true" class="icon icon icon-help"><use href="#icon-help"></svg>
                    Proceed with caution
                </div>
            </div>
        """

        self.assertHTMLEqual(expected, Template(template).render(Context()))

    def test_kwargs_with_filters(self):
        template = """
            {% load wagtailadmin_tags %}
            {% help_block status="warning"|upper %}Proceed with caution{% endhelp_block %}
        """

        expected = """
            <div class="help-block help-WARNING">
                <svg aria-hidden="true" class="icon icon icon-warning"><use href="#icon-warning"></svg>
                Proceed with caution
            </div>
        """

        self.assertHTMLEqual(expected, Template(template).render(Context()))

    def test_render_as_variable(self):
        template = """
            {% load wagtailadmin_tags %}
            {% help_block status="info" as help %}Proceed with caution{% endhelp_block %}
            <template>{{ help }}</template>
        """

        expected = """
            <template>
                <div class="help-block help-info">
                    <svg aria-hidden="true" class="icon icon icon-help"><use href="#icon-help"></svg>
                    Proceed with caution
                </div>
            </template>
        """

        self.assertHTMLEqual(expected, Template(template).render(Context()))


class FragmentTagTest(SimpleTestCase):
    def test_basic(self):
        context = Context({})

        template = """
            {% load wagtailadmin_tags %}
            {% fragment as my_fragment %}
            <p>Hello, World</p>
            {% endfragment %}
            Text coming after:
            {{ my_fragment }}
        """

        expected = """
            Text coming after:
            <p>Hello, World</p>
        """

        self.assertHTMLEqual(expected, Template(template).render(context))

    @override_settings(DEBUG=True)
    def test_syntax_error(self):
        template = """
            {% load wagtailadmin_tags %}
            {% fragment %}
            <p>Hello, World</p>
            {% endfragment %}
        """

        with self.assertRaises(TemplateSyntaxError):
            Template(template).render(Context())

    def test_with_variables(self):
        context = Context({"name": "jonathan wells"})

        template = """
            {% load wagtailadmin_tags %}
            {% fragment as my_fragment %}
                <p>Hello, {{ name|title }}</p>
            {% endfragment %}
            Text coming after:
            {{ my_fragment }}
        """

        expected = """
            Text coming after:
            <p>Hello, Jonathan Wells</p>
        """

        self.assertHTMLEqual(expected, Template(template).render(context))


class ClassnamesTagTest(SimpleTestCase):
    def test_with_single_arg(self):
        template = """
            {% load wagtailadmin_tags %}
            <p class="{% classnames "w-header" classname  %}">Hello!</p>
        """

        expected = """
            <p class="w-header">Hello!</p>
        """

        actual = Template(template).render(Context())

        self.assertHTMLEqual(expected, actual)

    def test_with_multiple_args(self):
        template = """
            {% load wagtailadmin_tags %}
            <p class="{% classnames "w-header" classname "w-header--merged" "w-header--hasform" %}">
                Hello!
            </p>
        """

        expected = """
            <p class="w-header w-header--merged w-header--hasform">
                Hello!
            </p>
        """

        actual = Template(template).render(Context())

        self.assertHTMLEqual(expected, actual)

    def test_with_falsy_args(self):
        template = """
            {% load wagtailadmin_tags %}
            <p class="{% classnames "w-header" classname "" %}">Hello!</p>
        """

        expected = """
            <p class="w-header">Hello!</p>
        """

        actual = Template(template).render(Context())

        self.assertEqual(expected.strip(), actual.strip())

    def test_with_args_with_extra_whitespace(self):
        context = Context(
            {
                "merged": "w-header--merged ",
                "search_form": " w-header--hasform",
                "name": " wagtail ",
            }
        )

        template = """
            {% load wagtailadmin_tags %}
            <p class="{% classnames "w-header" classname merged search_form name %}">Hello!</p>
        """

        expected = """
            <p class="w-header w-header--merged w-header--hasform wagtail">Hello!</p>
        """

        actual = Template(template).render(context)

        self.assertEqual(expected.strip(), actual.strip())

    def test_with_nested_lists(self):
        context = Context(
            {
                "nested": ["button-add", "button-base "],
                "has_falsey": ["", False, [], {}],
                "simple": " wagtail ",
            }
        )

        template = """
            {% load wagtailadmin_tags %}
            <button class="{% classnames nested "add-second " has_falsey simple %}">Hello!</button>
        """

        expected = """
            <button class="button-add button-base add-second wagtail">Hello!</button>
        """

        actual = Template(template).render(context)

        self.assertEqual(expected.strip(), actual.strip())


class IconTagTest(SimpleTestCase):
    def test_basic(self):
        template = """
            {% load wagtailadmin_tags %}
            {% icon name="cogs" %}
        """

        expected = """
            <svg aria-hidden="true" class="icon icon-cogs icon"><use href="#icon-cogs"></svg>
        """

        self.assertHTMLEqual(expected, Template(template).render(Context()))

    def test_with_classes_positional(self):
        template = """
            {% load wagtailadmin_tags %}
            {% icon "cogs" "myclass" %}
        """

        expected = """
            <svg aria-hidden="true" class="icon icon-cogs myclass"><use href="#icon-cogs"></svg>
        """

        self.assertHTMLEqual(expected, Template(template).render(Context()))

    def test_with_classes_keyword(self):
        template = """
            {% load wagtailadmin_tags %}
            {% icon name="warning" classname="myclass" %}
        """

        expected = """
            <svg aria-hidden="true" class="icon icon-warning myclass"><use href="#icon-warning"></svg>
        """

        self.assertHTMLEqual(expected, Template(template).render(Context()))


class StatusTagTest(SimpleTestCase):
    def test_render_block_component_span_variations(self):
        template = """
            {% load wagtailadmin_tags i18n %}
            {% status "live" classname="w-status--primary" %}
            {% status "live" %}
            {% trans "hidden translated label" as trans_hidden_label %}
            {% status "live" hidden_label=trans_hidden_label classname="w-status--primary" %}
            {% status %}
        """

        expected = """
            <span class="w-status w-status--primary">live</span>
            <span class="w-status">live</span>
            <span class="w-status w-status--primary"><span class="w-sr-only">hidden translated label</span>live</span>
            <span class="w-status"></span>
        """

        self.assertHTMLEqual(expected, Template(template).render(Context()))

    def test_render_block_component_anchor_variations(self):
        template = """
            {% load wagtailadmin_tags i18n %}
            {% trans "title" as trans_title %}
            {% trans "hidden label" as trans_hidden_label %}
            {% status "live" url="/test-url/" title=trans_title hidden_label=trans_hidden_label classname="w-status--primary" attrs='target="_blank" rel="noreferrer"' %}
            {% status "live" url="/test-url/" title=trans_title classname="w-status--primary" %}
            {% status "live" url="/test-url/" title=trans_title %}
            {% status  url="/test-url/" title=trans_title attrs='id="my-status"' %}
        """

        expected = """
            <a href="/test-url/" class="w-status w-status--primary" title="title" target="_blank" rel="noreferrer">
                <span class="w-sr-only">hidden label</span>
                live
            </a>
            <a href="/test-url/" class="w-status w-status--primary" title="title">
                live
            </a>
            <a href="/test-url/" class="w-status" title="title">
                live
            </a>
            <a href="/test-url/" class="w-status" title="title" id="my-status">
            </a>
        """

        self.assertHTMLEqual(expected, Template(template).render(Context()))

    def test_render_as_fragment(self):
        template = """
            {% load wagtailadmin_tags i18n %}
            {% fragment as var %}
                {% trans "title" as trans_title %}
                {% trans "hidden label" as trans_hidden_label %}
                {% status "live" url="/test-url/" title=trans_title hidden_label=trans_hidden_label classname="w-status--primary" attrs='target="_blank" rel="noreferrer"' %}
                {% status "live" hidden_label=trans_hidden_label classname="w-status--primary" attrs="data-example='present'" %}
            {% endfragment %}
            {{var}}
        """

        expected = """
            <a href="/test-url/" class="w-status w-status--primary" title="title" target="_blank" rel="noreferrer">
                <span class="w-sr-only">hidden label</span>
                live
            </a>
            <span class="w-status w-status--primary" data-example='present'>
                <span class="w-sr-only">hidden label</span>
                live
            </span>
        """

        self.assertHTMLEqual(expected, Template(template).render(Context()))


class BreadcrumbsTagTest(AdminTemplateTestUtils, WagtailTestUtils, SimpleTestCase):
    base_breadcrumb_items = []
    template = """
        {% load wagtailadmin_tags %}
        {% breadcrumbs items %}
    """

    def test_single_item(self):
        items = [{"label": "Something", "url": "/admin/something/"}]
        rendered = Template(self.template).render(Context({"items": items}))
        self.assertBreadcrumbsItemsRendered(items, rendered)
        # Without specifying is_expanded=False, the breadcrumbs should not be
        # collapsible anyway, so it is not controlled by Stimulus
        soup = self.get_soup(rendered)
        breadcrumbs = soup.select_one(".w-breadcrumbs")
        self.assertIsNotNone(breadcrumbs)
        self.assertIsNone(breadcrumbs.get("data-controller"))

    def test_trailing_no_url(self):
        items = [
            {"label": "Snippets", "url": "/admin/snippets/"},
            {"label": "People", "url": "/admin/snippets/people/"},
            {"label": "New: Person"},
        ]
        rendered = Template(self.template).render(Context({"items": items}))
        self.assertBreadcrumbsItemsRendered(items, rendered)

    def test_not_is_expanded(self):
        items = [
            {"label": "Snippets", "url": "/admin/snippets/"},
            {"label": "People", "url": "/admin/snippets/people/"},
            {"label": "Muddy Waters", "url": "/admin/snippets/people/1/edit/"},
        ]
        rendered = Template(self.template).render(Context({"items": items}))
        self.assertBreadcrumbsItemsRendered(items, rendered)

        soup = self.get_soup(rendered)
        controller = soup.select_one('[data-controller="w-breadcrumbs"]')
        toggle_button = soup.select_one('button[data-w-breadcrumbs-target="toggle"]')
        self.assertIsNotNone(controller)
        self.assertIsNotNone(toggle_button)
        # If is_expanded=False (the default), the breadcrumbs should be
        # collapsible via Stimulus
        soup = self.get_soup(rendered)
        breadcrumbs = soup.select_one(".w-breadcrumbs")
        self.assertIsNotNone(breadcrumbs)
        self.assertEqual(breadcrumbs.get("data-controller"), "w-breadcrumbs")

    def test_is_expanded(self):
        template = """
            {% load wagtailadmin_tags %}
            {% breadcrumbs items is_expanded=True %}
        """
        items = [
            {"label": "Snippets", "url": "/admin/snippets/"},
            {"label": "People", "url": "/admin/snippets/people/"},
            {"label": "Muddy Waters", "url": "/admin/snippets/people/1/edit/"},
        ]
        rendered = Template(template).render(Context({"items": items}))
        self.assertBreadcrumbsItemsRendered(items, rendered)

        soup = self.get_soup(rendered)
        controller = soup.select_one('[data-controller="w-breadcrumbs"]')
        toggle_button = soup.select_one('button[data-w-breadcrumbs-target="toggle"]')
        self.assertIsNone(controller)
        self.assertIsNone(toggle_button)
        # If is_expanded=True, the breadcrumbs should not be collapsible, so it
        # is not controlled by Stimulus
        soup = self.get_soup(rendered)
        breadcrumbs = soup.select_one(".w-breadcrumbs")
        self.assertIsNotNone(breadcrumbs)
        self.assertIsNone(breadcrumbs.get("data-controller"))

    def test_classname(self):
        template = """
            {% load wagtailadmin_tags %}
            {% breadcrumbs items classname="my-class" %}
        """
        items = [{"label": "Home", "url": "/admin/"}]
        rendered = Template(template).render(Context({"items": items}))
        self.assertBreadcrumbsItemsRendered(items, rendered)

        soup = self.get_soup(rendered)
        div = soup.select_one("div.w-breadcrumbs")
        self.assertIsNotNone(div)
        self.assertIn("my-class", div["class"])

    def test_icon_name(self):
        template = """
            {% load wagtailadmin_tags %}
            {% breadcrumbs items icon_name="site" %}
        """
        items = [
            {"label": "Home", "url": "/admin/"},
            {"label": "Something", "url": "/admin/something/"},
        ]
        rendered = Template(template).render(Context({"items": items}))
        self.assertBreadcrumbsItemsRendered(items, rendered)

        soup = self.get_soup(rendered)
        invalid_icons = soup.select("ol li:not(:last-child) svg use[href='#icon-site']")
        self.assertEqual(len(invalid_icons), 0)
        icon = soup.select_one("ol li:last-child svg use[href='#icon-site']")
        self.assertIsNotNone(icon)


class PageBreadcrumbsTagTest(AdminTemplateTestUtils, WagtailTestUtils, TestCase):
    fixtures = ["test.json"]
    base_breadcrumb_items = []

    def setUp(self):
        self.request = get_dummy_request()
        self.user = self.login()
        self.request.user = self.user

    def test_root_single_item(self):
        template = """
            {% load wagtailadmin_tags %}
            {% page_breadcrumbs page 'wagtailadmin_explore' url_root_name='wagtailadmin_explore_root' %}
        """
        page = Page.objects.get(id=1)
        items = [{"label": "Root", "url": "/admin/pages/"}]
        rendered = Template(template).render(
            Context({"page": page, "request": self.request})
        )
        self.assertBreadcrumbsItemsRendered(items, rendered)
        # Without specifying is_expanded=False, the breadcrumbs should not be
        # collapsible anyway, so it is not controlled by Stimulus
        soup = self.get_soup(rendered)
        breadcrumbs = soup.select_one(".w-breadcrumbs")
        self.assertIsNotNone(breadcrumbs)
        self.assertIsNone(breadcrumbs.get("data-controller"))

    def test_url_name(self):
        template = """
            {% load wagtailadmin_tags %}
            {% page_breadcrumbs page 'wagtailadmin_choose_page_child' %}
        """
        page = Page.objects.get(id=15)
        items = [
            {
                "label": "Root",
                "url": "/admin/choose-page/1/",
            },
            {
                "label": "Welcome to the Wagtail test site!",
                "url": "/admin/choose-page/2/",
            },
            {
                "label": "Events",
                "url": "/admin/choose-page/3/",
            },
            {
                "label": "Businessy events",
                "url": "/admin/choose-page/15/",
            },
        ]
        rendered = Template(template).render(
            Context({"page": page, "request": self.request})
        )
        self.assertBreadcrumbsItemsRendered(items, rendered)

    def test_not_include_self(self):
        template = """
            {% load wagtailadmin_tags %}
            {% page_breadcrumbs page 'wagtailadmin_explore' url_root_name='wagtailadmin_explore_root' include_self=False %}
        """
        page = Page.objects.get(id=15)
        items = [
            {"label": "Root", "url": "/admin/pages/"},
            {"label": "Welcome to the Wagtail test site!", "url": "/admin/pages/2/"},
            {"label": "Events", "url": "/admin/pages/3/"},
        ]
        rendered = Template(template).render(
            Context({"page": page, "request": self.request})
        )
        self.assertBreadcrumbsItemsRendered(items, rendered)

    def test_not_is_expanded(self):
        template = """
            {% load wagtailadmin_tags %}
            {% page_breadcrumbs page 'wagtailadmin_explore' url_root_name='wagtailadmin_explore_root' %}
        """
        page = Page.objects.get(id=15)
        items = [
            {"label": "Root", "url": "/admin/pages/"},
            {"label": "Welcome to the Wagtail test site!", "url": "/admin/pages/2/"},
            {"label": "Events", "url": "/admin/pages/3/"},
            {"label": "Businessy events", "url": "/admin/pages/15/"},
        ]
        rendered = Template(template).render(
            Context({"page": page, "request": self.request})
        )
        self.assertBreadcrumbsItemsRendered(items, rendered)
        # If is_expanded=False, the breadcrumbs should be collapsible via Stimulus
        soup = self.get_soup(rendered)
        breadcrumbs = soup.select_one(".w-breadcrumbs")
        self.assertIsNotNone(breadcrumbs)
        self.assertEqual(breadcrumbs.get("data-controller"), "w-breadcrumbs")

    def test_is_expanded(self):
        template = """
            {% load wagtailadmin_tags %}
            {% page_breadcrumbs page 'wagtailadmin_explore' url_root_name='wagtailadmin_explore_root' is_expanded=True %}
        """
        page = Page.objects.get(id=15)
        items = [
            {"label": "Root", "url": "/admin/pages/"},
            {"label": "Welcome to the Wagtail test site!", "url": "/admin/pages/2/"},
            {"label": "Events", "url": "/admin/pages/3/"},
            {"label": "Businessy events", "url": "/admin/pages/15/"},
        ]
        rendered = Template(template).render(
            Context({"page": page, "request": self.request})
        )
        self.assertBreadcrumbsItemsRendered(items, rendered)
        # If is_expanded=True, the breadcrumbs should not be collapsible, so it
        # is not controlled by Stimulus
        soup = self.get_soup(rendered)
        breadcrumbs = soup.select_one(".w-breadcrumbs")
        self.assertIsNotNone(breadcrumbs)
        self.assertIsNone(breadcrumbs.get("data-controller"))

    def test_querystring_value(self):
        template = """
            {% load wagtailadmin_tags %}
            {% page_breadcrumbs page 'wagtailadmin_explore' url_root_name='wagtailadmin_explore_root' querystring_value='?site=2&has_child_pages=true' %}
        """
        page = Page.objects.get(id=15)
        params = "?site=2&has_child_pages=true"
        items = [
            {"label": "Root", "url": f"/admin/pages/{params}"},
            {
                "label": "Welcome to the Wagtail test site!",
                "url": f"/admin/pages/2/{params}",
            },
            {"label": "Events", "url": f"/admin/pages/3/{params}"},
            {"label": "Businessy events", "url": f"/admin/pages/15/{params}"},
        ]
        rendered = Template(template).render(
            Context({"page": page, "request": self.request})
        )
        self.assertBreadcrumbsItemsRendered(items, rendered)

    def test_trailing_breadcrumb_title(self):
        template = """
            {% load wagtailadmin_tags %}
            {% page_breadcrumbs page 'wagtailadmin_explore' url_root_name='wagtailadmin_explore_root' trailing_breadcrumb_title='New: Simple Page' %}
        """
        page = Page.objects.get(id=15)
        items = [
            {"label": "Root", "url": "/admin/pages/"},
            {"label": "Welcome to the Wagtail test site!", "url": "/admin/pages/2/"},
            {"label": "Events", "url": "/admin/pages/3/"},
            {"label": "Businessy events", "url": "/admin/pages/15/"},
            {"label": "New: Simple Page"},
        ]
        rendered = Template(template).render(
            Context({"page": page, "request": self.request})
        )
        self.assertBreadcrumbsItemsRendered(items, rendered)

    def test_classname(self):
        template = """
            {% load wagtailadmin_tags %}
            {% page_breadcrumbs page 'wagtailadmin_choose_page_child' classname='my-class' %}
        """
        page = Page.objects.get(id=1)
        items = [{"label": "Root", "url": "/admin/choose-page/1/"}]
        rendered = Template(template).render(
            Context({"page": page, "request": self.request})
        )
        self.assertBreadcrumbsItemsRendered(items, rendered)

        soup = self.get_soup(rendered)
        div = soup.select_one("div.w-breadcrumbs")
        self.assertIsNotNone(div)
        self.assertIn("my-class", div["class"])

    def test_icon_name(self):
        template = """
            {% load wagtailadmin_tags %}
            {% page_breadcrumbs page 'wagtailadmin_explore' icon_name='site' %}
        """
        page = Page.objects.get(id=3)
        items = [
            {"label": "Root", "url": "/admin/pages/1/"},
            {"label": "Welcome to the Wagtail test site!", "url": "/admin/pages/2/"},
            {"label": "Events", "url": "/admin/pages/3/"},
        ]
        rendered = Template(template).render(
            Context({"page": page, "request": self.request})
        )
        self.assertBreadcrumbsItemsRendered(items, rendered)

        soup = self.get_soup(rendered)
        invalid_icons = soup.select("ol li:not(:last-child) svg use[href='#icon-site']")
        self.assertEqual(len(invalid_icons), 0)
        icon = soup.select_one("ol li:last-child svg use[href='#icon-site']")
        self.assertIsNotNone(icon)


class ThemeColorSchemeTest(AdminTemplateTestUtils, WagtailTestUtils, TestCase):
    def setUp(self):
        self.request = get_dummy_request()
        self.user = self.login()
        self.request.user = self.user
        self.profile = UserProfile.get_for_user(self.user)

    def test_default_mode(self):
        template = """
            {% load wagtailadmin_tags %}
            <meta name="color-scheme" content="{% admin_theme_color_scheme %}">
        """
        rendered = Template(template).render(Context({"request": self.request}))

        soup = self.get_soup(rendered)
        meta_tag = soup.find("meta", {"name": "color-scheme"})
        self.assertIsNotNone(meta_tag)
        self.assertEqual(meta_tag["content"], "dark light")

    def test_dark_mode(self):
        self.profile.theme = "dark"
        self.profile.save()

        template = """
            {% load wagtailadmin_tags %}
            <meta name="color-scheme" content="{% admin_theme_color_scheme %}">
        """
        rendered = Template(template).render(Context({"request": self.request}))

        soup = self.get_soup(rendered)
        meta_tag = soup.find("meta", {"name": "color-scheme"})
        self.assertIsNotNone(meta_tag)
        self.assertEqual(meta_tag["content"], "dark")

    def test_light_mode(self):
        self.profile.theme = "light"
        self.profile.save()

        template = """
            {% load wagtailadmin_tags %}
            <meta name="color-scheme" content="{% admin_theme_color_scheme %}">
        """
        rendered = Template(template).render(Context({"request": self.request}))

        soup = self.get_soup(rendered)
        meta_tag = soup.find("meta", {"name": "color-scheme"})
        self.assertIsNotNone(meta_tag)
        self.assertEqual(meta_tag["content"], "light")



import logging
import uuid

from django.core.exceptions import PermissionDenied

from wagtail.log_actions import log
from wagtail.models.copying import _copy, _copy_m2m_relations
from wagtail.models.i18n import TranslatableMixin

logger = logging.getLogger("wagtail")


class CreatePageAliasIntegrityError(RuntimeError):
    """
    Raised when creating an alias of a page cannot be performed for data integrity reasons.
    """

    pass


class CreatePageAliasPermissionError(PermissionDenied):
    """
    Raised when creating an alias of a page cannot be performed due to insufficient permissions.
    """

    pass


class CreatePageAliasAction:
    """
    Creates an alias of the given page.

    An alias is like a copy, but an alias remains in sync with the original page. They
    are not directly editable and do not have revisions.

    You can convert an alias into a regular page by setting the .alias_of attribute to None
    and creating an initial revision.

    :param recursive: create aliases of the page's subtree, defaults to False
    :type recursive: boolean, optional
    :param parent: The page to create the new alias under
    :type parent: Page, optional
    :param update_slug: The slug of the new alias page, defaults to the slug of the original page
    :type update_slug: string, optional
    :param update_locale: The locale of the new alias page, defaults to the locale of the original page
    :type update_locale: Locale, optional
    :param user: The user who is performing this action. This user would be assigned as the owner of the new page and appear in the audit log
    :type user: User, optional
    :param log_action: Override the log action with a custom one. or pass None to skip logging, defaults to 'wagtail.create_alias'
    :type log_action: string or None, optional
    :param reset_translation_key: Generate new translation_keys for the page and any translatable child objects, defaults to False
    :type reset_translation_key: boolean, optional
    """

    def __init__(
        self,
        page,
        *,
        recursive=False,
        parent=None,
        update_slug=None,
        update_locale=None,
        user=None,
        log_action="wagtail.create_alias",
        reset_translation_key=True,
        _mpnode_attrs=None,
    ):
        self.page = page
        self.recursive = recursive
        self.parent = parent
        self.update_slug = update_slug
        self.update_locale = update_locale
        self.user = user
        self.log_action = log_action
        self.reset_translation_key = reset_translation_key
        self._mpnode_attrs = _mpnode_attrs

    def check(self, skip_permission_checks=False):
        parent = self.parent or self.page.get_parent()
        if self.recursive and (
            parent == self.page or parent.is_descendant_of(self.page)
        ):
            raise CreatePageAliasIntegrityError(
                "You cannot copy a tree branch recursively into itself"
            )

        if (
            self.user
            and not skip_permission_checks
            and not parent.permissions_for_user(self.user).can_publish_subpage()
        ):
            raise CreatePageAliasPermissionError(
                "You do not have permission to publish a page at the destination"
            )

    def _create_alias(
        self,
        page,
        *,
        recursive,
        parent,
        update_slug,
        update_locale,
        user,
        log_action,
        reset_translation_key,
        _mpnode_attrs,
    ):
        specific_page = page.specific

        # FIXME: Switch to the same fields that are excluded from copy
        # We can't do this right now because we can't exclude fields from with_content_json
        # which we use for updating aliases
        exclude_fields = [
            "id",
            "path",
            "depth",
            "numchild",
            "url_path",
            "path",
            "index_entries",
            "postgres_index_entries",
            "latest_revision",  # for page aliases do not have revisions
        ]

        update_attrs = {
            "alias_of": page,
            # Aliases don't have revisions so the draft title should always match the live title
            "draft_title": page.title,
            # Likewise, an alias page can't have unpublished changes if it's live
            "has_unpublished_changes": not page.live,
        }

        if update_slug:
            update_attrs["slug"] = update_slug

        if update_locale:
            update_attrs["locale"] = update_locale

        if user:
            update_attrs["owner"] = user

        # When we're not copying for translation, we should give the translation_key a new value
        if reset_translation_key:
            update_attrs["translation_key"] = uuid.uuid4()

        alias, child_object_map = _copy(
            specific_page, update_attrs=update_attrs, exclude_fields=exclude_fields
        )

        # Update any translatable child objects
        for child_object in child_object_map.values():
            if isinstance(child_object, TranslatableMixin):
                if update_locale:
                    child_object.locale = update_locale

                # When we're not copying for translation,
                # we should give the translation_key a new value for each child object as well.
                if reset_translation_key:
                    child_object.translation_key = uuid.uuid4()

        # Save the new page
        if _mpnode_attrs:
            # We've got a tree position already reserved. Perform a quick save.
            alias.path = _mpnode_attrs[0]
            alias.depth = _mpnode_attrs[1]
            alias.save(clean=False)

        else:
            if parent:
                alias = parent.add_child(instance=alias)
            else:
                alias = page.add_sibling(instance=alias)

            _mpnode_attrs = (alias.path, alias.depth)

        _copy_m2m_relations(specific_page, alias, exclude_fields=exclude_fields)

        # Log
        if log_action:
            source_parent = specific_page.get_parent()
            log(
                instance=alias,
                action=log_action,
                user=user,
                data={
                    "page": {"id": alias.id, "title": alias.get_admin_display_title()},
                    "source": {
                        "id": source_parent.id,
                        "title": source_parent.specific_deferred.get_admin_display_title(),
                    }
                    if source_parent
                    else None,
                    "destination": {
                        "id": parent.id,
                        "title": parent.specific_deferred.get_admin_display_title(),
                    }
                    if parent
                    else None,
                },
            )

        logger.info(
            'Page alias created: "%s" id=%d from=%d', alias.title, alias.id, page.id
        )

        from wagtail.models import Page, PageViewRestriction

        # Copy child pages
        if recursive:
            numchild = 0

            for child_page in page.get_children().specific().iterator():
                newdepth = _mpnode_attrs[1] + 1
                child_mpnode_attrs = (
                    Page._get_path(_mpnode_attrs[0], newdepth, numchild),
                    newdepth,
                )
                numchild += 1
                self._create_alias(
                    child_page,
                    recursive=True,
                    parent=alias,
                    update_slug=None,
                    update_locale=update_locale,
                    user=user,
                    log_action=log_action,
                    reset_translation_key=reset_translation_key,
                    _mpnode_attrs=child_mpnode_attrs,
                )

            if numchild > 0:
                alias.numchild = numchild
                alias.save(clean=False, update_fields=["numchild"])

        # Copy across any view restrictions defined directly on the page,
        # unless the destination page already has view restrictions defined
        if parent:
            parent_page_restriction = parent.get_view_restrictions()
        else:
            parent_page_restriction = page.get_parent().get_view_restrictions()

        if not parent_page_restriction.exists():
            for view_restriction in page.view_restrictions.all():
                view_restriction_copy = PageViewRestriction(
                    restriction_type=view_restriction.restriction_type,
                    password=view_restriction.password,
                    page=alias,
                )
                view_restriction_copy.save(user=self.user)
                view_restriction_copy.groups.set(view_restriction.groups.all())

        return alias

    def execute(self, skip_permission_checks=False):
        self.check(skip_permission_checks=skip_permission_checks)

        return self._create_alias(
            self.page,
            recursive=self.recursive,
            parent=self.parent,
            update_slug=self.update_slug,
            update_locale=self.update_locale,
            user=self.user,
            log_action=self.log_action,
            reset_translation_key=self.reset_translation_key,
            _mpnode_attrs=self._mpnode_attrs,
        )



# Generated by Django 1.9.4 on 2016-03-29 04:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("snippetstests", "0005_multisectionrichtextsnippet_richtextsection"),
    ]

    operations = [
        migrations.CreateModel(
            name="StandardSnippetWithCustomPrimaryKey",
            fields=[
                (
                    "snippet_id",
                    models.CharField(max_length=255, primary_key=True, serialize=False),
                ),
                ("text", models.CharField(max_length=255)),
            ],
        ),
    ]



from django.contrib.admin.utils import quote
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from wagtail.admin.views.bulk_action.mixins import ReferenceIndexMixin
from wagtail.admin.views.generic import BeforeAfterHookMixin
from wagtail.snippets.bulk_actions.snippet_bulk_action import SnippetBulkAction
from wagtail.snippets.permissions import get_permission_name


class DeleteBulkAction(BeforeAfterHookMixin, ReferenceIndexMixin, SnippetBulkAction):
    display_name = _("Delete")
    action_type = "delete"
    aria_label = _("Delete selected snippets")
    template_name = "wagtailsnippets/bulk_actions/confirm_bulk_delete.html"
    action_priority = 30
    classes = {"serious"}

    @cached_property
    def _actionable_objects(self):
        # Cache the actionable objects so that we can use them early in the
        # dispatch() method, and reuse it later when needed. We don't name this
        # property "actionable_objects" because the method returns a tuple of
        # `(objects, {"items_with_no_access": items_with_no_access})` and the base
        # class assigns the `objects` to self.actionable_objects.
        return super().get_actionable_objects()

    def get_actionable_objects(self):
        return self._actionable_objects

    # Run the snippet deletion hooks that we've had since before bulk actions
    # were introduced.
    # These are different from the base bulk action hooks that are run by
    # __run_{before,after}_hooks(), as those are only called in form_valid().
    # Here, we make use of the BeforeAfterHookMixin to run the before hook at
    # the beginning of dispatch(), and the after hook at the end of form_valid(),
    # which is how the delete-multiple view for snippets used to work.

    def run_before_hook(self):
        # The method returns a tuple of
        # `(objects, {"items_with_no_access": items_with_no_access})`, hence [0]
        objects = self.get_actionable_objects()[0]
        return self.run_hook("before_delete_snippet", self.request, objects)

    def run_after_hook(self):
        objects = self.get_actionable_objects()[0]
        return self.run_hook("after_delete_snippet", self.request, objects)

    def check_perm(self, snippet):
        if getattr(self, "can_delete_items", None) is None:
            # since snippets permissions are not enforced per object, makes sense to just check once per model request
            self.can_delete_items = self.request.user.has_perm(
                get_permission_name("delete", self.model)
            )
        return self.can_delete_items

    @classmethod
    def execute_action(cls, objects, user=None, **kwargs):
        kwargs["self"].model.objects.filter(
            pk__in=[snippet.pk for snippet in objects]
        ).delete()
        return len(objects), 0

    def get_usage_url(self, snippet):
        return (
            reverse(
                snippet.snippet_viewset.get_url_name("usage"),
                args=(quote(snippet.pk),),
            )
            + "?describe_on_delete=1"
        )

    def get_success_message(self, num_parent_objects, num_child_objects):
        if num_parent_objects == 1:
            return _("%(model_name)s '%(object)s' deleted.") % {
                "model_name": capfirst(self.model._meta.verbose_name),
                "object": self.actionable_objects[0],
            }
        else:
            return ngettext(
                "%(count)d %(model_name)s deleted.",
                "%(count)d %(model_name)s deleted.",
                num_parent_objects,
            ) % {
                "model_name": capfirst(self.model._meta.verbose_name_plural),
                "count": num_parent_objects,
            }



from django.forms import Media
from django.utils.functional import cached_property

from wagtail.admin.telepath import register as register_telepath_adapter

from .base import Panel


class PanelGroup(Panel):
    """
    Abstract class for panels that manage a set of sub-panels.
    Concrete subclasses must attach a 'children' property
    """

    def __init__(self, children=(), *args, **kwargs):
        permission = kwargs.pop("permission", None)
        super().__init__(*args, **kwargs)
        self.children = children
        self.permission = permission

    def clone_kwargs(self):
        kwargs = super().clone_kwargs()
        kwargs["children"] = self.children
        kwargs["permission"] = self.permission
        return kwargs

    def get_form_options(self):
        if self.model is None:
            raise AttributeError(
                "%s is not bound to a model yet. Use `.bind_to_model(model)` "
                "before using this method." % self.__class__.__name__
            )

        options = {}

        # Merge in form options from each child in turn, combining values that are types that we
        # know how to combine (i.e. lists, dicts and sets)
        for child in self.children:
            child_options = child.get_form_options()
            for key, new_val in child_options.items():
                if key not in options:
                    # if val is a known mutable container type that we're going to merge subsequent
                    # child values into, create a copy so that we don't risk that change leaking
                    # back into the child's internal state
                    if (
                        isinstance(new_val, list)
                        or isinstance(new_val, dict)
                        or isinstance(new_val, set)
                    ):
                        options[key] = new_val.copy()
                    else:
                        options[key] = new_val
                else:
                    current_val = options[key]
                    if isinstance(current_val, list) and isinstance(
                        new_val, (list, tuple)
                    ):
                        current_val.extend(new_val)
                    elif isinstance(current_val, tuple) and isinstance(
                        new_val, (list, tuple)
                    ):
                        options[key] = list(current_val).extend(new_val)
                    elif isinstance(current_val, dict) and isinstance(new_val, dict):
                        current_val.update(new_val)
                    elif isinstance(current_val, set) and isinstance(new_val, set):
                        current_val.update(new_val)
                    else:
                        raise ValueError(
                            "Don't know how to merge values %r and %r for form option %r"
                            % (current_val, new_val, key)
                        )

        return options

    def on_model_bound(self):
        from .model_utils import expand_panel_list

        child_panels = expand_panel_list(self.model, self.children)
        self.children = [child.bind_to_model(self.model) for child in child_panels]

    @cached_property
    def child_identifiers(self):
        """
        A list of identifiers corresponding to child panels in ``self.children``, formed from the clean_name property
        but validated to be unique and non-empty.
        """
        used_names = set()
        result = []
        for panel in self.children:
            base_name = panel.clean_name or "panel"
            candidate_name = base_name
            suffix = 0
            while candidate_name in used_names:
                suffix += 1
                candidate_name = "%s%d" % (base_name, suffix)

            result.append(candidate_name)
            used_names.add(candidate_name)

        return result

    @register_telepath_adapter
    class BoundPanel(Panel.BoundPanel):
        @cached_property
        def children(self):
            return [
                child.get_bound_panel(
                    instance=self.instance,
                    request=self.request,
                    form=self.form,
                    prefix=(f"{self.prefix}-child-{identifier}"),
                )
                for child, identifier in zip(
                    self.panel.children, self.panel.child_identifiers
                )
            ]

        @cached_property
        def visible_children(self):
            return [child for child in self.children if child.is_shown()]

        @cached_property
        def visible_children_with_identifiers(self):
            return [
                (child, identifier)
                for child, identifier in zip(
                    self.children, self.panel.child_identifiers
                )
                if child.is_shown()
            ]

        def show_panel_furniture(self):
            return any(child.show_panel_furniture() for child in self.children)

        def is_shown(self):
            """
            Check permissions on the panel group overall then check if any children
            are shown.
            """

            if self.panel.permission:
                if not self.request.user.has_perm(self.panel.permission):
                    return False

            return any(child.is_shown() for child in self.children)

        @property
        def media(self):
            media = Media()
            for item in self.visible_children:
                media += item.media
            return media

        def get_comparison(self):
            comparators = []

            for child in self.children:
                comparators.extend(child.get_comparison())

            return comparators

        telepath_adapter_name = "wagtail.panels.PanelGroup"

        def js_opts(self):
            opts = super().js_opts()
            opts["children"] = self.visible_children
            return opts


class TabbedInterface(PanelGroup):
    BASE_ATTRS = {
        "data-controller": "w-tabs",
        "data-action": "popstate@window->w-tabs#select",
        "data-w-tabs-use-location-value": "true",
    }

    class BoundPanel(PanelGroup.BoundPanel):
        template_name = "wagtailadmin/panels/tabbed_interface.html"


class ObjectList(PanelGroup):
    class BoundPanel(PanelGroup.BoundPanel):
        template_name = "wagtailadmin/panels/object_list.html"


class FieldRowPanel(PanelGroup):
    class BoundPanel(PanelGroup.BoundPanel):
        template_name = "wagtailadmin/panels/field_row_panel.html"


class MultiFieldPanel(PanelGroup):
    class BoundPanel(PanelGroup.BoundPanel):
        template_name = "wagtailadmin/panels/multi_field_panel.html"



# Placeholder for panel types defined in wagtail.admin.panels.
# These are needed because we wish to define properties such as `content_panels` on core models
# such as Page, but importing from wagtail.admin would create a circular import. We therefore use a
# placeholder object, and swap it out for the real panel class inside
# `wagtail.admin.panels.model_utils.expand_panel_list` at the same time as converting strings to
# FieldPanel instances.

from django.conf import settings
from django.utils.functional import cached_property
from django.utils.module_loading import import_string


class PanelPlaceholder:
    def __init__(self, path, args, kwargs):
        self.path = path
        self.args = args
        self.kwargs = kwargs

    @cached_property
    def panel_class(self):
        return import_string(self.path)

    def construct(self):
        return self.panel_class(*self.args, **self.kwargs)


class CommentPanelPlaceholder(PanelPlaceholder):
    def __init__(self):
        super().__init__(
            "wagtail.admin.panels.CommentPanel",
            [],
            {},
        )

    def construct(self):
        if getattr(settings, "WAGTAILADMIN_COMMENTS_ENABLED", True):
            return super().construct()



from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand

from wagtail.models import ReferenceIndex


def model_name(model):
    return f"{model.__module__}.{model.__name__}"


class Command(BaseCommand):
    def handle(self, **options):
        self.stdout.write("Reference index entries:")
        object_count = 0

        for model in sorted(apps.get_models(), key=lambda m: model_name(m)):
            if not ReferenceIndex.is_indexed(model):
                continue

            content_types = [
                ContentType.objects.get_for_model(
                    model_or_object, for_concrete_model=False
                )
                for model_or_object in ([model] + model._meta.get_parent_list())
            ]
            content_type = content_types[0]
            base_content_type = content_types[-1]

            count = ReferenceIndex.objects.filter(
                content_type=content_type, base_content_type=base_content_type
            ).count()
            self.stdout.write(f"{count:>6}  {model_name(model)}")
            object_count += count

        self.stdout.write(f"Total entries: {object_count}")



from wagtail.management.commands.publish_scheduled import (
    Command as PublishScheduledCommand,
)


class Command(PublishScheduledCommand):
    """
    Alias for the publish_scheduled management command for backwards-compatibility.
    """



import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0020_add_index_on_page_first_published_at"),
    ]

    operations = [
        migrations.AlterField(
            model_name="grouppagepermission",
            name="group",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                related_name="page_permissions",
                to="auth.Group",
                verbose_name="group",
            ),
        ),
        migrations.AlterField(
            model_name="grouppagepermission",
            name="page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                related_name="group_permissions",
                to="wagtailcore.Page",
                verbose_name="page",
            ),
        ),
        migrations.AlterField(
            model_name="grouppagepermission",
            name="permission_type",
            field=models.CharField(
                max_length=20,
                verbose_name="permission type",
                choices=[
                    ("add", "Add/edit pages you own"),
                    ("edit", "Edit any page"),
                    ("publish", "Publish any page"),
                    ("lock", "Lock/unlock any page"),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="content_type",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                related_name="pages",
                to="contenttypes.ContentType",
                verbose_name="content type",
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="expire_at",
            field=models.DateTimeField(
                null=True,
                blank=True,
                help_text="Please add a date-time in the form YYYY-MM-DD hh:mm.",
                verbose_name="expiry date/time",
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="expired",
            field=models.BooleanField(
                default=False, editable=False, verbose_name="expired"
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="first_published_at",
            field=models.DateTimeField(
                null=True,
                db_index=True,
                editable=False,
                verbose_name="first published at",
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="go_live_at",
            field=models.DateTimeField(
                null=True,
                blank=True,
                help_text="Please add a date-time in the form YYYY-MM-DD hh:mm.",
                verbose_name="go live date/time",
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="has_unpublished_changes",
            field=models.BooleanField(
                default=False, editable=False, verbose_name="has unpublished changes"
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="latest_revision_created_at",
            field=models.DateTimeField(
                null=True, editable=False, verbose_name="latest revision created at"
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="live",
            field=models.BooleanField(
                default=True, editable=False, verbose_name="live"
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="locked",
            field=models.BooleanField(
                default=False, editable=False, verbose_name="locked"
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="owner",
            field=models.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="owned_pages",
                null=True,
                to=settings.AUTH_USER_MODEL,
                editable=False,
                verbose_name="owner",
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="search_description",
            field=models.TextField(blank=True, verbose_name="search description"),
        ),
        migrations.AlterField(
            model_name="page",
            name="seo_title",
            field=models.CharField(
                max_length=255,
                blank=True,
                help_text="Optional. 'Search Engine Friendly' title. This will appear at the top of the browser window.",
                verbose_name="page title",
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="show_in_menus",
            field=models.BooleanField(
                default=False,
                help_text="Whether a link to this page will appear in automatically generated menus",
                verbose_name="show in menus",
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="slug",
            field=models.SlugField(
                max_length=255,
                verbose_name="slug",
                help_text="The name of the page as it will appear in URLs e.g http://domain.com/blog/[my-slug]/",
            ),
        ),
        migrations.AlterField(
            model_name="page",
            name="title",
            field=models.CharField(
                max_length=255,
                help_text="The page title as you'd like it to be seen by the public",
                verbose_name="title",
            ),
        ),
        migrations.AlterField(
            model_name="pagerevision",
            name="approved_go_live_at",
            field=models.DateTimeField(
                null=True, blank=True, verbose_name="approved go live at"
            ),
        ),
        migrations.AlterField(
            model_name="pagerevision",
            name="content_json",
            field=models.TextField(verbose_name="content JSON"),
        ),
        migrations.AlterField(
            model_name="pagerevision",
            name="created_at",
            field=models.DateTimeField(verbose_name="created at"),
        ),
        migrations.AlterField(
            model_name="pagerevision",
            name="page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                related_name="revisions",
                to="wagtailcore.Page",
                verbose_name="page",
            ),
        ),
        migrations.AlterField(
            model_name="pagerevision",
            name="submitted_for_moderation",
            field=models.BooleanField(
                default=False, db_index=True, verbose_name="submitted for moderation"
            ),
        ),
        migrations.AlterField(
            model_name="pagerevision",
            name="user",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                blank=True,
                null=True,
                to=settings.AUTH_USER_MODEL,
                verbose_name="user",
            ),
        ),
        migrations.AlterField(
            model_name="pageviewrestriction",
            name="page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                related_name="view_restrictions",
                to="wagtailcore.Page",
                verbose_name="page",
            ),
        ),
        migrations.AlterField(
            model_name="pageviewrestriction",
            name="password",
            field=models.CharField(max_length=255, verbose_name="password"),
        ),
        migrations.AlterField(
            model_name="site",
            name="hostname",
            field=models.CharField(
                max_length=255, db_index=True, verbose_name="hostname"
            ),
        ),
        migrations.AlterField(
            model_name="site",
            name="is_default_site",
            field=models.BooleanField(
                default=False,
                help_text="If true, this site will handle requests for all other hostnames that do not have a site entry of their own",
                verbose_name="is default site",
            ),
        ),
        migrations.AlterField(
            model_name="site",
            name="port",
            field=models.IntegerField(
                default=80,
                help_text="Set this to something other than 80 if you need a specific port number to appear in URLs (e.g. development on port 8000). Does not affect request handling (so port forwarding still works).",
                verbose_name="port",
            ),
        ),
        migrations.AlterField(
            model_name="site",
            name="root_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                related_name="sites_rooted_here",
                to="wagtailcore.Page",
                verbose_name="root page",
            ),
        ),
    ]



from django.utils.translation import gettext_lazy

from wagtail.admin.forms.pages import WagtailAdminPageForm
from wagtail.models import Page
from wagtail.utils.decorators import cached_classmethod

from .group import ObjectList, TabbedInterface

Page.base_form_class = WagtailAdminPageForm


@cached_classmethod
def _get_page_edit_handler(cls):
    """
    Get the panel to use in the Wagtail admin when editing this page type.
    """
    if hasattr(cls, "edit_handler"):
        edit_handler = cls.edit_handler
    else:
        # construct a TabbedInterface made up of content_panels, promote_panels
        # and settings_panels, skipping any which are empty
        tabs = []

        if cls.content_panels:
            tabs.append(ObjectList(cls.content_panels, heading=gettext_lazy("Content")))
        if cls.promote_panels:
            tabs.append(ObjectList(cls.promote_panels, heading=gettext_lazy("Promote")))
        if cls.settings_panels:
            tabs.append(
                ObjectList(cls.settings_panels, heading=gettext_lazy("Settings"))
            )

        edit_handler = TabbedInterface(tabs, base_form_class=cls.base_form_class)

    return edit_handler.bind_to_model(cls)


Page.get_edit_handler = _get_page_edit_handler



from django.core import checks
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from wagtail.actions.publish_revision import PublishRevisionAction
from wagtail.actions.unpublish import UnpublishAction
from wagtail.locks import ScheduledForPublishLock

from .revisions import RevisionMixin


class DraftStateMixin(models.Model):
    live = models.BooleanField(verbose_name=_("live"), default=True, editable=False)
    has_unpublished_changes = models.BooleanField(
        verbose_name=_("has unpublished changes"), default=False, editable=False
    )

    first_published_at = models.DateTimeField(
        verbose_name=_("first published at"), blank=True, null=True, db_index=True
    )
    last_published_at = models.DateTimeField(
        verbose_name=_("last published at"), null=True, editable=False
    )
    live_revision = models.ForeignKey(
        "wagtailcore.Revision",
        related_name="+",
        verbose_name=_("live revision"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        editable=False,
    )

    go_live_at = models.DateTimeField(
        verbose_name=_("go live date/time"), blank=True, null=True
    )
    expire_at = models.DateTimeField(
        verbose_name=_("expiry date/time"), blank=True, null=True
    )
    expired = models.BooleanField(
        verbose_name=_("expired"), default=False, editable=False
    )

    class Meta:
        abstract = True

    @classmethod
    def check(cls, **kwargs):
        return [
            *super().check(**kwargs),
            *cls._check_revision_mixin(),
        ]

    @classmethod
    def _check_revision_mixin(cls):
        mro = cls.mro()
        error = checks.Error(
            "DraftStateMixin requires RevisionMixin to be applied after DraftStateMixin.",
            hint="Add RevisionMixin to the model's base classes after DraftStateMixin.",
            obj=cls,
            id="wagtailcore.E004",
        )

        try:
            if mro.index(RevisionMixin) < mro.index(DraftStateMixin):
                return [error]
        except ValueError:
            return [error]

        return []

    @property
    def approved_schedule(self):
        return self.scheduled_revision is not None

    @property
    def status_string(self):
        if not self.live:
            if self.expired:
                return _("expired")
            elif self.approved_schedule:
                return _("scheduled")
            else:
                return _("draft")
        else:
            if self.approved_schedule:
                return _("live + scheduled")
            elif self.has_unpublished_changes:
                return _("live + draft")
            else:
                return _("live")

    def publish(
        self,
        revision,
        user=None,
        changed=True,
        log_action=True,
        previous_revision=None,
        skip_permission_checks=False,
    ):
        """
        Publish a revision of the object by applying the changes in the revision to the live object.

        :param revision: Revision to publish.
        :type revision: Revision
        :param user: The publishing user.
        :param changed: Indicated whether content has changed.
        :param log_action: Flag for the logging action, pass ``False`` to skip logging.
        :param previous_revision: Indicates a revision reversal. Should be set to the previous revision instance.
        :type previous_revision: Revision
        """
        return PublishRevisionAction(
            revision,
            user=user,
            changed=changed,
            log_action=log_action,
            previous_revision=previous_revision,
        ).execute(skip_permission_checks=skip_permission_checks)

    def unpublish(self, set_expired=False, commit=True, user=None, log_action=True):
        """
        Unpublish the live object.

        :param set_expired: Mark the object as expired.
        :param commit: Commit the changes to the database.
        :param user: The unpublishing user.
        :param log_action: Flag for the logging action, pass ``False`` to skip logging.
        """
        return UnpublishAction(
            self,
            set_expired=set_expired,
            commit=commit,
            user=user,
            log_action=log_action,
        ).execute()

    def with_content_json(self, content):
        """
        Similar to :meth:`RevisionMixin.with_content_json`,
        but with the following fields also preserved:

        * ``live``
        * ``has_unpublished_changes``
        * ``first_published_at``
        """
        obj = super().with_content_json(content)

        # Ensure other values that are meaningful for the object as a whole (rather than
        # to a specific revision) are preserved
        obj.live = self.live
        obj.has_unpublished_changes = self.has_unpublished_changes
        obj.first_published_at = self.first_published_at

        return obj

    def get_latest_revision_as_object(self):
        if not self.has_unpublished_changes:
            # Use the live database copy in preference to the revision record, as:
            # 1) this will pick up any changes that have been made directly to the model,
            #    such as automated data imports;
            # 2) it ensures that inline child objects pick up real database IDs even if
            #    those are absent from the revision data. (If this wasn't the case, the child
            #    objects would be recreated with new IDs on next publish - see #1853)
            return self

        latest_revision = self.get_latest_revision()

        if latest_revision:
            return latest_revision.as_object()
        else:
            return self

    @cached_property
    def scheduled_revision(self):
        return self.revisions.filter(approved_go_live_at__isnull=False).first()

    def get_scheduled_revision_as_object(self):
        scheduled_revision = self.scheduled_revision
        return scheduled_revision and scheduled_revision.as_object()

    def _update_from_revision(self, revision, changed=True):
        update_fields = ["latest_revision"]
        self.latest_revision = revision

        if changed:
            self.has_unpublished_changes = True
            update_fields.append("has_unpublished_changes")

        self.save(update_fields=update_fields)

    def get_lock(self):
        # Scheduled publishing lock should take precedence over other locks
        if self.approved_schedule:
            return ScheduledForPublishLock(self)
        return super().get_lock()



import datetime
from collections import defaultdict

import django_filters
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.db.models import IntegerField, Value
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from wagtail.admin.admin_url_finder import AdminURLFinder
from wagtail.admin.filters import (
    ContentTypeFilter,
)
from wagtail.admin.views.pages.history import PageHistoryFilterSet
from wagtail.coreutils import get_content_type_label
from wagtail.log_actions import registry as log_action_registry
from wagtail.models import PageLogEntry

from .base import ReportView


def get_users_for_filter(user):
    user_ids = set()
    for log_model in log_action_registry.get_log_entry_models():
        user_ids.update(log_model.objects.viewable_by_user(user).get_user_ids())

    User = get_user_model()
    return User.objects.filter(pk__in=user_ids).order_by(User.USERNAME_FIELD)


def get_content_types_for_filter(user):
    content_type_ids = set()
    for log_model in log_action_registry.get_log_entry_models():
        content_type_ids.update(
            log_model.objects.viewable_by_user(user).get_content_type_ids()
        )

    return ContentType.objects.filter(pk__in=content_type_ids).order_by("model")


def get_actions_for_filter(user):
    # Only return those actions used by log entries visible to the user.
    actions = set()
    for log_model in log_action_registry.get_log_entry_models():
        actions.update(log_model.objects.viewable_by_user(user).get_actions())

    return [
        action for action in log_action_registry.get_choices() if action[0] in actions
    ]


class SiteHistoryReportFilterSet(PageHistoryFilterSet):
    label = django_filters.CharFilter(label=_("Name"), lookup_expr="icontains")
    object_type = ContentTypeFilter(
        label=_("Type"),
        method="filter_object_type",
        queryset=lambda request: get_content_types_for_filter(request.user),
    )

    def filter_object_type(self, queryset, name, value):
        return queryset.filter_on_content_type(value)

    def get_action_choices(self):
        return get_actions_for_filter(self.request.user)

    def get_user_choices(self):
        return get_users_for_filter(self.request.user)

    class Meta:
        model = PageLogEntry
        fields = []


class LogEntriesView(ReportView):
    results_template_name = "wagtailadmin/reports/site_history_results.html"
    page_title = _("Site history")
    header_icon = "history"
    filterset_class = SiteHistoryReportFilterSet
    index_url_name = "wagtailadmin_reports:site_history"
    index_results_url_name = "wagtailadmin_reports:site_history_results"

    export_headings = {
        "object_id": _("ID"),
        "label": _("Name"),
        "content_type": _("Type"),
        "action": _("Action type"),
        "user_display_name": _("User"),
        "timestamp": _("Date/Time"),
    }
    list_export = [
        "object_id",
        "label",
        "content_type",
        "action",
        "user_display_name",
        "timestamp",
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.custom_field_preprocess = self.custom_field_preprocess.copy()
        self.custom_field_preprocess["action"] = {
            self.FORMAT_CSV: self.get_action_label,
            self.FORMAT_XLSX: self.get_action_label,
        }
        self.custom_field_preprocess["content_type"] = {
            self.FORMAT_CSV: get_content_type_label,
            self.FORMAT_XLSX: get_content_type_label,
        }

    def get_filename(self):
        return "audit-log-{}".format(datetime.datetime.today().strftime("%Y-%m-%d"))

    def get_filtered_queryset(self):
        """
        Since this report combines records from multiple log models, the standard pattern of
        returning a queryset from get_queryset() to be filtered by filter_queryset() is not
        possible - the subquery for each log model must be filtered separately before joining
        with union().

        Additionally, a union() on standard model-based querysets will return a queryset based on
        the first model in the union, so instances of the other model(s) would be returned as the
        wrong type. To avoid this, we construct values() querysets as follows:

        1. For each model, construct a values() queryset consisting of id, timestamp and an
           annotation to indicate which model it is, and filter this with filter_queryset
        2. Form a union() queryset from these queries, and order it by -timestamp
           (this is the result returned from get_filtered_queryset)
        3. Apply pagination (done in MultipleObjectMixin.get_context_data)
        4. (In decorate_paginated_queryset:) For each model included in the result set, look up
           the set of model instances by ID. Use these to form a final list of model instances
           in the same order as the query.
        """
        queryset = None

        # Retrieve the set of registered log models, and cast it to a list so that we assign
        # an index number to each one; this index number will be used to distinguish models
        # in the combined results
        self.log_models = list(log_action_registry.get_log_entry_models())

        for log_model_index, log_model in enumerate(self.log_models):
            sub_queryset = (
                log_model.objects.viewable_by_user(self.request.user)
                .values("pk", "timestamp")
                .annotate(
                    log_model_index=Value(log_model_index, output_field=IntegerField())
                )
            )
            sub_queryset = self.filter_queryset(sub_queryset)
            # disable any native ordering on the queryset; we will re-apply it on the combined result
            sub_queryset = sub_queryset.order_by()
            if queryset is None:
                queryset = sub_queryset
            else:
                queryset = queryset.union(sub_queryset)

        return queryset.order_by("-timestamp")

    def decorate_paginated_queryset(self, queryset):
        # build lists of ids from queryset, grouped by log model index
        pks_by_log_model_index = defaultdict(list)
        for row in queryset:
            pks_by_log_model_index[row["log_model_index"]].append(row["pk"])

        url_finder = AdminURLFinder(self.request.user)

        # for each log model found in the queryset, look up the set of log entries by id
        # and build a lookup table
        object_lookup = {}
        for log_model_index, pks in pks_by_log_model_index.items():
            log_entries = (
                self.log_models[log_model_index]
                .objects.prefetch_related("user__wagtail_userprofile", "content_type")
                .filter(pk__in=pks)
                .with_instances()
            )
            for log_entry, instance in log_entries:
                # annotate log entry with an 'edit_url' property
                log_entry.edit_url = url_finder.get_edit_url(instance)
                object_lookup[(log_model_index, log_entry.pk)] = log_entry

        # return items from our lookup table in the order of the original queryset
        return [object_lookup[(row["log_model_index"], row["pk"])] for row in queryset]

    def get_action_label(self, action):
        return force_str(log_action_registry.get_action_label(action))



from django.db import migrations


def add_page_lock_permission_to_moderators(apps, schema_editor):
    Group = apps.get_model("auth.Group")
    Page = apps.get_model("wagtailcore.Page")
    GroupPagePermission = apps.get_model("wagtailcore.GroupPagePermission")

    root_pages = Page.objects.filter(depth=1)

    try:
        moderators_group = Group.objects.get(name="Moderators")

        for page in root_pages:
            GroupPagePermission.objects.create(
                group=moderators_group, page=page, permission_type="lock"
            )

    except Group.DoesNotExist:
        pass


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0004_page_locked"),
    ]

    operations = [
        migrations.RunPython(add_page_lock_permission_to_moderators),
    ]



# Generated by Django 4.2.7 on 2024-02-01 21:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0091_remove_revision_submitted_for_moderation'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionviewrestriction',
            name='password',
            field=models.CharField(blank=True, help_text='Shared passwords should not be used to protect sensitive content. Anyone who has this password will be able to view the content.', max_length=255, verbose_name='shared password'),
        ),
        migrations.AlterField(
            model_name='collectionviewrestriction',
            name='restriction_type',
            field=models.CharField(choices=[('none', 'Public'), ('password', 'Private, accessible with a shared password'), ('login', 'Private, accessible to any logged-in users'), ('groups', 'Private, accessible to users in specific groups')], max_length=20),
        ),
        migrations.AlterField(
            model_name='pageviewrestriction',
            name='password',
            field=models.CharField(blank=True, help_text='Shared passwords should not be used to protect sensitive content. Anyone who has this password will be able to view the content.', max_length=255, verbose_name='shared password'),
        ),
        migrations.AlterField(
            model_name='pageviewrestriction',
            name='restriction_type',
            field=models.CharField(choices=[('none', 'Public'), ('password', 'Private, accessible with a shared password'), ('login', 'Private, accessible to any logged-in users'), ('groups', 'Private, accessible to users in specific groups')], max_length=20),
        ),
    ]



from django.forms import Media, widgets
from django.utils.safestring import mark_safe


class CustomRichTextArea(widgets.Textarea):
    def render(self, name, value, attrs=None, renderer=None):
        # mock rendering for individual custom widget

        return mark_safe(
            '<template data-controller="custom-editor" data-id="{}">{}</template>'.format(
                attrs["id"],
                super().render(name, value, attrs),
            )
        )

    @property
    def media(self):
        return Media(js=["vendor/custom_editor.js"])


class LegacyRichTextArea(widgets.Textarea):
    def render(self, name, value, attrs=None, renderer=None):
        # mock rendering for individual custom widget
        return mark_safe(
            '<template data-controller="legacy-editor" data-id="{}">{}</template>'.format(
                attrs["id"],
                super().render(name, value, attrs),
            )
        )

    @property
    def media(self):
        return Media(js=["vendor/legacy_editor.js"])



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0001_initial"),
        ("wagtailcore", "0025_collection_initial_data"),
    ]

    operations = [
        migrations.CreateModel(
            name="GroupCollectionPermission",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        primary_key=True,
                        auto_created=True,
                        verbose_name="ID",
                    ),
                ),
                (
                    "collection",
                    models.ForeignKey(
                        related_name="group_permissions",
                        verbose_name="collection",
                        to="wagtailcore.Collection",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        related_name="collection_permissions",
                        verbose_name="group",
                        to="auth.Group",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "permission",
                    models.ForeignKey(
                        to="auth.Permission",
                        verbose_name="permission",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={
                "verbose_name": "group collection permission",
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name="groupcollectionpermission",
            unique_together={("group", "collection", "permission")},
        ),
    ]



from django.contrib.auth.models import AbstractBaseUser, Group
from django.test import SimpleTestCase, TestCase
from django.urls import reverse
from django.utils.http import urlencode

from wagtail import hooks
from wagtail.admin import widgets as wagtailadmin_widgets
from wagtail.admin.wagtail_hooks import page_header_buttons, page_listing_more_buttons
from wagtail.admin.widgets.button import Button
from wagtail.models import Page
from wagtail.test.testapp.models import SimplePage
from wagtail.test.utils import WagtailTestUtils
from wagtail.utils.deprecation import RemovedInWagtail80Warning


class TestButtonsHooks(WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()

        self.root_page = Page.objects.get(id=2)
        self.child_page = self.root_page.add_child(
            instance=SimplePage(
                title="Public page",
                content="hello",
                live=True,
            )
        )


class TestPageListingButtonsHooks(TestButtonsHooks):
    def test_register_page_listing_buttons_deprecated_class(self):
        def page_listing_buttons_deprecated_class(page, user, next_url=None):
            yield wagtailadmin_widgets.PageListingButton(
                "Another useless page listing button", "/custom-url", priority=10
            )

        with (
            hooks.register_temporarily(
                "register_page_listing_buttons", page_listing_buttons_deprecated_class
            ),
            self.assertWarnsMessage(
                RemovedInWagtail80Warning,
                "`PageListingButton` is deprecated. "
                "Use `wagtail.admin.widgets.button.ListingButton` instead.",
            ),
        ):
            response = self.client.get(
                reverse("wagtailadmin_explore", args=(self.root_page.id,))
            )

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtailadmin/shared/button_with_dropdown.html"
        )
        self.assertTemplateUsed(
            response, "wagtailadmin/pages/listing/_button_with_dropdown.html"
        )
        self.assertTemplateUsed(response, "wagtailadmin/shared/buttons.html")

        self.assertContains(response, "Another useless page listing button")

    def test_register_page_listing_buttons_new_signature(self):
        def page_listing_buttons_new_signature(page, user, next_url=None):
            if not isinstance(user, AbstractBaseUser):
                raise TypeError("expected a user instance")

            yield wagtailadmin_widgets.ListingButton(
                "Another useless page listing button", "/custom-url", priority=10
            )

        with hooks.register_temporarily(
            "register_page_listing_buttons", page_listing_buttons_new_signature
        ):
            response = self.client.get(
                reverse("wagtailadmin_explore", args=(self.root_page.id,))
            )

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtailadmin/shared/button_with_dropdown.html"
        )
        self.assertTemplateUsed(
            response, "wagtailadmin/pages/listing/_button_with_dropdown.html"
        )
        self.assertTemplateUsed(response, "wagtailadmin/shared/buttons.html")

        self.assertContains(response, "Another useless page listing button")


class TestPageListingMoreButtonsHooks(TestButtonsHooks):
    def test_register_page_listing_more_buttons_with_new_signature(self):
        def page_listing_more_buttons(page, user, next_url=None):
            if not isinstance(user, AbstractBaseUser):
                raise TypeError("expected a user instance")

            yield wagtailadmin_widgets.Button(
                'Another useless button in default "More" dropdown',
                "/custom-url",
                priority=10,
            )

        with hooks.register_temporarily(
            "register_page_listing_more_buttons", page_listing_more_buttons
        ):
            response = self.client.get(
                reverse("wagtailadmin_explore", args=(self.root_page.id,))
            )

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtailadmin/shared/button_with_dropdown.html"
        )
        self.assertTemplateUsed(
            response, "wagtailadmin/pages/listing/_button_with_dropdown.html"
        )
        self.assertTemplateUsed(response, "wagtailadmin/shared/buttons.html")

        self.assertContains(
            response, "Another useless button in default &quot;More&quot; dropdown"
        )

    def test_custom_button_with_dropdown_with_new_signature(self):
        def page_custom_listing_buttons(page, user, next_url=None):
            yield wagtailadmin_widgets.ButtonWithDropdownFromHook(
                "One more more button",
                hook_name="register_page_listing_one_more_more_buttons",
                page=page,
                user=user,
                next_url=next_url,
                attrs={"target": "_blank", "rel": "noreferrer"},
                priority=50,
            )

        def page_custom_listing_more_buttons(page, user, next_url=None):
            if not isinstance(user, AbstractBaseUser):
                raise TypeError("expected a user instance")

            yield wagtailadmin_widgets.Button(
                'Another useless dropdown button in "One more more button" dropdown',
                "/custom-url",
                priority=10,
            )

        with (
            hooks.register_temporarily(
                "register_page_listing_buttons",
                page_custom_listing_buttons,
            ),
            hooks.register_temporarily(
                "register_page_listing_one_more_more_buttons",
                page_custom_listing_more_buttons,
            ),
        ):
            response = self.client.get(
                reverse("wagtailadmin_explore", args=(self.root_page.id,))
            )

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtailadmin/shared/button_with_dropdown.html"
        )
        self.assertTemplateUsed(
            response, "wagtailadmin/pages/listing/_button_with_dropdown.html"
        )
        self.assertTemplateUsed(response, "wagtailadmin/shared/buttons.html")

        self.assertContains(response, "One more more button")
        self.assertContains(
            response,
            "Another useless dropdown button in &quot;One more more button&quot; dropdown",
        )

    def test_delete_button_with_next_url(self):
        """
        Ensure that the built in delete button supports a next_url provided.
        """

        # page_listing_more_button generator yields only `Delete button` with this permission set
        page = self.root_page
        base_url = reverse("wagtailadmin_pages:delete", args=[page.id])

        next_url = "a/random/url/"
        full_url = base_url + "?" + urlencode({"next": next_url})

        buttons = page_listing_more_buttons(page, user=self.user, next_url=next_url)
        delete_button = next(button for button in buttons if button.label == "Delete")

        self.assertEqual(delete_button.url, full_url)

    def test_buttons_with_rel_attribute(self):
        """
        Ensure that PageMenuItem can specify a rel attribute for the link.
        """
        self.child_page.save_revision()
        response = self.client.get(
            reverse("wagtailadmin_explore", args=(self.root_page.id,))
        )
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)
        buttons = soup.select("td li [data-controller='w-dropdown'] a[rel]")
        self.assertEqual(len(buttons), 2)
        self.assertEqual(
            {button.text.strip() for button in buttons},
            {"View live", "View draft"},
        )
        self.assertEqual(
            [button.get("rel") for button in buttons],
            [["noreferrer"], ["noreferrer"]],
        )

    def test_delete_button_with_invalid_next_url(self):
        """
        Ensure that the built in delete button on page listing will not use
        the next_url provided if that URL is directing the user to edit the page.
        As the page is now deleted and cannot be edited.
        """

        page = self.root_page

        base_url = reverse("wagtailadmin_pages:delete", args=[page.id])
        next_url = reverse("wagtailadmin_explore", args=[page.id])

        buttons = page_listing_more_buttons(page, user=self.user, next_url=next_url)

        delete_button = next(button for button in buttons if button.label == "Delete")

        # check that the next_url is NOT included as it will not be available after deletion
        self.assertEqual(delete_button.url, base_url)

        # check that the unpublish button does correctly still include the next_url
        unpublish_base_url = reverse("wagtailadmin_pages:unpublish", args=[page.id])
        unpublish_button = next(
            button for button in buttons if button.label == "Unpublish"
        )
        full_url = unpublish_base_url + "?" + urlencode({"next": next_url})
        self.assertEqual(unpublish_button.url, full_url)

    def test_reorder_button_visibility(self):
        page = self.root_page

        # Test with a user with no publish permission (and thus no ability to reorder)
        editor = self.create_user(username="editor", password="password")
        editor.groups.add(Group.objects.get(name="Editors"))

        # no button returned
        buttons = [
            button
            for button in page_listing_more_buttons(page, user=editor)
            if button.is_shown(user=editor)
        ]
        self.assertEqual(
            len([button for button in buttons if button.label == "Sort menu order"]), 0
        )

        # Test with a user with publish permission
        publisher = self.create_user(username="publisher", password="password")
        publisher.groups.add(Group.objects.get(name="Moderators"))

        # page_listing_more_button generator yields `Sort menu order button`
        buttons = [
            button
            for button in page_listing_more_buttons(page, user=publisher)
            if button.is_shown(user=publisher)
        ]
        reorder_button = next(
            button for button in buttons if button.label == "Sort menu order"
        )

        self.assertEqual(reorder_button.url, "/admin/pages/%d/?ordering=ord" % page.id)


class TestPageHeaderButtonsHooks(TestButtonsHooks):
    def test_register_page_header_buttons_new_signature(self):
        def custom_page_header_buttons(page, user, view_name, next_url=None):
            if not isinstance(user, AbstractBaseUser):
                raise TypeError("expected a user instance")

            if view_name != "edit":
                raise ValueError("expected view_name to be 'edit'")

            yield wagtailadmin_widgets.Button(
                "Another useless header button", "/custom-url", priority=10
            )

        with hooks.register_temporarily(
            "register_page_header_buttons", custom_page_header_buttons
        ):
            response = self.client.get(
                reverse("wagtailadmin_pages:edit", args=(self.root_page.id,))
            )

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtailadmin/pages/listing/_page_header_buttons.html"
        )

        self.assertContains(response, "Another useless header button")

    def test_delete_button_with_next_url(self):
        """
        Ensure that the built in delete button supports a next_url provided.
        """

        page = self.root_page
        base_url = reverse("wagtailadmin_pages:delete", args=[page.id])

        next_url = "a/random/url/"
        full_url = base_url + "?" + urlencode({"next": next_url})

        buttons = page_header_buttons(
            page, self.user, view_name="index", next_url=next_url
        )
        delete_button = next(button for button in buttons if button.label == "Delete")

        self.assertEqual(delete_button.url, full_url)

    def test_delete_button_with_invalid_next_url(self):
        """
        Ensure that the built in delete button on page edit/home (header button) will not use
        the next_url provided if that URL is directing the user to edit the page.
        As the page is now deleted and cannot be edited.
        """

        page = self.root_page

        base_url = reverse("wagtailadmin_pages:delete", args=[page.id])
        next_url = reverse("wagtailadmin_explore", args=[page.id])

        buttons = page_header_buttons(
            page, self.user, view_name="index", next_url=next_url
        )

        delete_button = next(button for button in buttons if button.label == "Delete")

        # check that the next_url is NOT included as it will not be available after deletion (page listing)
        self.assertEqual(delete_button.url, base_url)

        base_url = reverse("wagtailadmin_pages:delete", args=[page.id])
        next_url = reverse("wagtailadmin_pages:edit", args=[page.id])

        buttons = page_header_buttons(
            page, self.user, view_name="index", next_url=next_url
        )

        delete_button = next(button for button in buttons if button.label == "Delete")

        # check that the next_url is NOT included as it will not be available after deletion (edit page)
        self.assertEqual(delete_button.url, base_url)

        # check that any buttons after do correctly still include the next_url
        unpublish_base_url = reverse("wagtailadmin_pages:unpublish", args=[page.id])
        unpublish_button = next(
            button for button in buttons if button.label == "Unpublish"
        )
        full_url = unpublish_base_url + "?" + urlencode({"next": next_url})
        self.assertEqual(unpublish_button.url, full_url)


class ButtonComparisonTestCase(SimpleTestCase):
    """Tests the comparison functions."""

    def setUp(self):
        self.button1 = Button(
            "Label 1", "/url1", classname="class1 class2", priority=100
        )
        self.button2 = Button(
            "Label 2", "/url2", classname="class2 class3", priority=200
        )
        self.button3 = Button(
            "Label 1", "/url3", classname="class1 class2", priority=300
        )
        self.button4 = Button(
            "Label 1", "/url1", classname="class1 class2", priority=100
        )

    def test_eq(self):
        # Same properties, should be equal
        self.assertTrue(self.button1 == self.button4)

        # Different priority, should not be equal
        self.assertFalse(self.button1 == self.button2)

        # Different URL, should not be equal
        self.assertFalse(self.button1 == self.button3)

        # Not a Button, should not be equal
        self.assertFalse(self.button1 == "Something")

    def test_lt(self):
        # Less priority, should be True
        self.assertTrue(self.button1 < self.button2)

        # Same label, but less priority, should be True
        self.assertTrue(self.button1 < self.button3)

        # Greater priority, should be False
        self.assertFalse(self.button2 < self.button1)

        # Not a Button, should raise TypeError
        with self.assertRaises(TypeError):
            self.button1 < "Something"

    def test_le(self):
        # Less priority, should be True
        self.assertTrue(self.button1 <= self.button2)

        # Same label, but less priority, should be True
        self.assertTrue(self.button1 <= self.button3)

        # Same object, should be True
        self.assertTrue(self.button1 <= self.button1)

        # Same label and priority, should be True
        self.assertTrue(self.button1 <= self.button4)

        # Greater priority, should be False
        self.assertFalse(self.button2 <= self.button1)

        # Not a Button, should raise TypeError
        with self.assertRaises(TypeError):
            self.button1 <= "Something"

    def test_gt(self):
        # Greater priority, should be True
        self.assertTrue(self.button2 > self.button1)

        # Same label, but greater priority, should be True
        self.assertTrue(self.button3 > self.button1)

        # Less priority, should be False
        self.assertFalse(self.button1 > self.button2)

        # Not a Button, should raise TypeError
        with self.assertRaises(TypeError):
            self.button1 > "Something"

    def test_ge(self):
        # Greater priority, should be True
        self.assertTrue(self.button2 >= self.button1)

        # Same label, but greater priority, should be True
        self.assertTrue(self.button3 >= self.button1)

        # Same object, should be True
        self.assertTrue(self.button1 >= self.button1)

        # Same label and priority, should be True
        self.assertTrue(self.button1 >= self.button4)

        # Less priority, should be False
        self.assertFalse(self.button1 >= self.button2)



from django.apps import apps
from django.core.signals import setting_changed
from django.dispatch import receiver

from wagtail.models import Page

from .model_utils import get_edit_handler


@receiver(setting_changed)
def reset_edit_handler_cache(**kwargs):
    """
    Clear page edit handler cache when global WAGTAILADMIN_COMMENTS_ENABLED settings are changed
    """
    if kwargs["setting"] == "WAGTAILADMIN_COMMENTS_ENABLED":
        for model in apps.get_models():
            if issubclass(model, Page):
                model.get_edit_handler.cache_clear()
        get_edit_handler.cache_clear()



import json

from django.http import HttpResponseBadRequest, JsonResponse
from django.views import View

from wagtail.users.models import UserProfile


class DismissiblesView(View):
    def get(self, request, *args, **kwargs):
        # The UserProfile may not exist for the user, in which case return an empty object
        profile = getattr(request.user, "wagtail_userprofile", None)
        dismissibles = profile.dismissibles if profile else {}
        return JsonResponse(dismissibles)

    def patch(self, request, *args, **kwargs):
        try:
            updates = json.loads(request.body)
        except json.JSONDecodeError:
            return HttpResponseBadRequest()

        # Make sure the UserProfile exists
        profile = UserProfile.get_for_user(request.user)
        profile.dismissibles.update(updates)
        profile.save(update_fields=["dismissibles"])
        return JsonResponse(profile.dismissibles)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class WagtailRoutablePageAppConfig(AppConfig):
    name = "wagtail.contrib.routable_page"
    label = "wagtailroutablepage"
    verbose_name = _("Wagtail routablepage")



import unittest
from unittest import skip

from django.db import connection
from django.test.testcases import TransactionTestCase
from django.test.utils import override_settings

from wagtail.search.query import Not, PlainText
from wagtail.search.tests.test_backends import BackendTests
from wagtail.test.search import models


@unittest.skipUnless(connection.vendor == "mysql", "The current database is not MySQL")
@override_settings(
    WAGTAILSEARCH_BACKENDS={
        "default": {
            "BACKEND": "wagtail.search.backends.database.mysql.mysql",
        }
    }
)
class TestMySQLSearchBackend(BackendTests, TransactionTestCase):
    backend_path = "wagtail.search.backends.database.mysql.mysql"

    # Overrides parent method, because there's a slight difference in what the MySQL backend supports/accepts as search queries.
    def test_not(self):
        all_other_titles = {
            "A Clash of Kings",
            "A Game of Thrones",
            "A Storm of Swords",
            "Foundation",
            "Learning Python",
            "The Hobbit",
            "The Two Towers",
            "The Fellowship of the Ring",
            "The Return of the King",
            "The Rust Programming Language",
            "Two Scoops of Django 1.11",
            "Programming Rust",
        }

        results = self.backend.search(
            Not(PlainText("javascript")), models.Book.objects.all()
        )
        self.assertSetEqual({r.title for r in results}, all_other_titles)

        results = self.backend.search(
            ~PlainText("javascript"), models.Book.objects.all()
        )
        self.assertSetEqual({r.title for r in results}, all_other_titles)

        # Tests multiple words
        results = self.backend.search(
            ~PlainText("javascript the"), models.Book.objects.all()
        )
        # NOTE: The difference with the parent method is here. As we're querying NOT 'javascript the', all entries containing both words should be excluded, but MySQL doesn't index stopwords in FULLTEXT indexes by default, so the JavaScript books won't match the query, since the 'the' word is excluded from the index. Therefore, both books will get returned.
        self.assertSetEqual(
            {r.title for r in results},
            all_other_titles
            | {"JavaScript: The Definitive Guide", "JavaScript: The good parts"},
        )

        # Tests multiple words too, but this time the second word is not a stopword
        results = self.backend.search(
            ~PlainText("javascript parts"), models.Book.objects.all()
        )
        self.assertSetEqual(
            {r.title for r in results},
            all_other_titles | {"JavaScript: The Definitive Guide"},
        )

    def test_empty_search(self):
        results = self.backend.search("", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            set(),
        )

        results = self.backend.search(" ", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            set(),
        )

        results = self.backend.search("*", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            set(),
        )

    def test_empty_autocomplete(self):
        results = self.backend.autocomplete("", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            set(),
        )

        results = self.backend.autocomplete(" ", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            set(),
        )

        results = self.backend.autocomplete("*", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            set(),
        )

    def test_symbols_in_search_term(self):
        # symbols as their own tokens should be ignored
        results = self.backend.search("javascript @ parts", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            {"JavaScript: The good parts"},
        )

        results = self.backend.search("javascript parts @", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            {"JavaScript: The good parts"},
        )

        results = self.backend.search("@ javascript parts", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            {"JavaScript: The good parts"},
        )

        # tokens containing both symbols and alphanumerics should not be discarded
        # or treated as equivalent to the same token without symbols
        results = self.backend.search("java@script parts", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            set(),
        )

    def test_autocomplete_with_symbols(self):
        # the * is not part of the autocomplete mechanism, but if someone includes it
        # we want it to be gracefully ignored
        results = self.backend.autocomplete("parts javasc*", models.Book.objects.all())
        self.assertSetEqual(
            {r.title for r in results},
            {"JavaScript: The good parts"},
        )

    def test_lexeme_validation(self):
        from wagtail.search.backends.database.mysql.query import Lexeme

        with self.assertRaisesMessage(ValueError, "Lexeme value cannot be empty."):
            Lexeme("")

        Lexeme("hello")  # no error

        with self.assertRaisesMessage(
            ValueError,
            "Lexeme value 'hello world' must consist of alphanumeric characters "
            "and '_' only.",
        ):
            Lexeme("hello world")

        with self.assertRaisesMessage(
            ValueError,
            "Lexeme value 'rimsky@korsakov' must consist of alphanumeric characters "
            "and '_' only.",
        ):
            Lexeme("rimsky@korsakov")

    @skip(
        "The MySQL backend doesn't support choosing individual fields for the search, only (body, title) or (autocomplete) fields may be searched."
    )
    def test_search_on_individual_field(self):
        return super().test_search_on_individual_field()

    @skip("The MySQL backend doesn't support boosting.")
    def test_search_boosting_on_related_fields(self):
        return super().test_search_boosting_on_related_fields()

    @skip("The MySQL backend doesn't support boosting.")
    def test_boost(self):
        return super().test_boost()

    @skip("The MySQL backend doesn't score annotations.")
    def test_annotate_score(self):
        return super().test_annotate_score()

    @skip("The MySQL backend doesn't score annotations.")
    def test_annotate_score_with_slice(self):
        return super().test_annotate_score_with_slice()

    @skip("The MySQL backend doesn't guarantee correct ranking of results.")
    def test_ranking(self):
        return super().test_ranking()



from wagtail.models import Collection, Locale, Page, Site, Task, Workflow
from wagtail.permission_policies import ModelPermissionPolicy
from wagtail.permission_policies.collections import CollectionManagementPermissionPolicy
from wagtail.permission_policies.pages import PagePermissionPolicy

page_permission_policy = PagePermissionPolicy(Page)
site_permission_policy = ModelPermissionPolicy(Site)
collection_permission_policy = CollectionManagementPermissionPolicy(Collection)
task_permission_policy = ModelPermissionPolicy(Task)
workflow_permission_policy = ModelPermissionPolicy(Workflow)
locale_permission_policy = ModelPermissionPolicy(Locale)



# Generated by Django 3.1.8 on 2021-06-22 14:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailsearchpromotions", "0003_query_querydailyhits"),
    ]

    # Changed to a no-op in Wagtail 6.0.
    # This migration previously copied the contents of the Query table from wagtailsearch to
    # wagtailsearchpromotions; however, this can't be done after wagtailsearch migration 0008,
    # which removes the Query table (and we can't force it to run before, because the
    # searchpromotions app can be installed at any time). However, any project that needs this
    # data migration would have already applied the real version of this migration while they
    # were running Wagtail 5.
    operations = []



import datetime
from collections import OrderedDict

from django.contrib.admin.utils import quote
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.forms import CheckboxSelectMultiple
from django.http import Http404
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.functional import classproperty
from django.utils.translation import gettext, gettext_lazy, ngettext
from django.views.generic import TemplateView
from django_filters import DateFromToRangeFilter

from wagtail.admin import messages
from wagtail.admin.filters import (
    DateRangePickerWidget,
    MultipleContentTypeFilter,
    WagtailFilterSet,
)
from wagtail.admin.ui.tables import Column, TitleColumn
from wagtail.admin.utils import get_valid_next_url_from_request
from wagtail.admin.views.generic import PermissionCheckedMixin
from wagtail.admin.views.generic.base import BaseListingView
from wagtail.admin.views.mixins import SpreadsheetExportMixin
from wagtail.admin.views.pages.listing import PageFilterSet, PageListingMixin
from wagtail.contrib.forms.models import FormMixin
from wagtail.contrib.forms.utils import get_form_types, get_forms_for_user
from wagtail.models import Page
from wagtail.permissions import page_permission_policy


def get_submissions_list_view(request, *args, **kwargs):
    """Call the form page's list submissions view class"""
    page_id = kwargs.get("page_id")
    form_page = get_object_or_404(Page, id=page_id).specific
    if not isinstance(form_page, FormMixin):
        raise Http404
    return form_page.serve_submissions_list_view(request, *args, **kwargs)


class ContentTypeColumn(Column):
    edit_url_name = "wagtailadmin_pages:edit"
    cell_template_name = "wagtailforms/content_type_column.html"

    def get_url(self, instance):
        return reverse(self.edit_url_name, args=(quote(instance.pk),))

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["url"] = self.get_url(instance)
        return context


class FormPageFilterSet(PageFilterSet):
    content_type = MultipleContentTypeFilter(
        label=gettext_lazy("Page type"),
        queryset=lambda request: ContentType.objects.filter(
            pk__in=[ct.pk for ct in get_form_types()]
        ).order_by("model"),
        widget=CheckboxSelectMultiple,
    )


class FormPagesListView(PageListingMixin, PermissionCheckedMixin, BaseListingView):
    """Lists the available form pages for the current user"""

    permission_policy = page_permission_policy
    any_permission_required = {
        "add",
        "change",
        "publish",
        "bulk_delete",
        "lock",
        "unlock",
    }
    template_name = "wagtailforms/index.html"
    results_template_name = "wagtailforms/index_results.html"
    context_object_name = "form_pages"
    paginate_by = 20
    page_kwarg = "p"
    index_url_name = "wagtailforms:index"
    index_results_url_name = "wagtailforms:index_results"
    page_title = gettext_lazy("Forms")
    header_icon = "form"
    model = Page
    is_searchable = True
    filterset_class = FormPageFilterSet

    @classproperty
    def columns(self):
        columns = [
            col for col in PageListingMixin.columns if col.name not in {"title", "type"}
        ]
        columns.insert(
            1,
            TitleColumn(
                "title",
                classname="title",
                label=gettext_lazy("Title"),
                url_name="wagtailforms:list_submissions",
                sort_key="title",
                width="30%",
            ),
        )
        columns.insert(
            -1,
            ContentTypeColumn(
                "content_type",
                label=gettext_lazy("Origin"),
                sort_key="content_type",
                width="20%",
            ),
        )
        return columns

    def get_breadcrumbs_items(self):
        return self.breadcrumbs_items + [
            {"url": "", "label": self.page_title, "sublabel": gettext("Pages")},
        ]

    def get_base_queryset(self):
        """Return the queryset of form pages for this view"""
        pages = get_forms_for_user(self.request.user).select_related("content_type")
        return self.annotate_queryset(pages)


class DeleteSubmissionsView(TemplateView):
    """Delete the selected submissions"""

    template_name = "wagtailforms/confirm_delete.html"
    page = None
    submissions = None
    success_url_name = "wagtailforms:list_submissions"

    def get_queryset(self):
        """Returns a queryset for the selected submissions"""
        submission_ids = self.request.GET.getlist("selected-submissions")
        submission_class = self.page.get_submission_class()
        return submission_class._default_manager.filter(id__in=submission_ids)

    def handle_delete(self, submissions):
        """Deletes the given queryset"""
        count = submissions.count()
        submissions.delete()
        messages.success(
            self.request,
            ngettext(
                "One submission has been deleted.",
                "%(count)d submissions have been deleted.",
                count,
            )
            % {"count": count},
        )

    def get_success_url(self):
        """Returns the success URL to redirect to after a successful deletion"""
        next_url = get_valid_next_url_from_request(self.request)
        if next_url:
            return next_url
        return reverse(self.success_url_name, args=(self.page.id,))

    def dispatch(self, request, *args, **kwargs):
        """Check permissions, set the page and submissions, handle delete"""
        page_id = kwargs.get("page_id")

        if not get_forms_for_user(self.request.user).filter(id=page_id).exists():
            raise PermissionDenied

        self.page = get_object_or_404(Page, id=page_id).specific

        self.submissions = self.get_queryset()

        if self.request.method == "POST":
            self.handle_delete(self.submissions)
            return redirect(self.get_success_url())

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """Get the context for this view"""
        context = super().get_context_data(**kwargs)

        context.update(
            {
                "page": self.page,
                "submissions": self.submissions,
            }
        )
        context["next_url"] = self.get_success_url()

        return context


class SubmissionsListFilterSet(WagtailFilterSet):
    date = DateFromToRangeFilter(
        label=gettext_lazy("Submission date"),
        field_name="submit_time",
        widget=DateRangePickerWidget,
    )


class SubmissionsListView(SpreadsheetExportMixin, BaseListingView):
    """Lists submissions for the provided form page"""

    template_name = "wagtailforms/submissions_index.html"
    results_template_name = "wagtailforms/list_submissions.html"
    context_object_name = "submissions"
    form_page = None
    default_ordering = ("-submit_time",)
    ordering_csv = ("submit_time",)  # keep legacy CSV ordering
    orderable_fields = (
        "id",
        "submit_time",
    )  # used to validate ordering in URL
    page_title = gettext_lazy("Form data")
    header_icon = "form"
    paginate_by = 20
    filterset_class = SubmissionsListFilterSet
    forms_index_url_name = "wagtailforms:index"
    index_url_name = "wagtailforms:list_submissions"
    index_results_url_name = "wagtailforms:list_submissions_results"
    show_export_buttons = True

    def dispatch(self, request, *args, **kwargs):
        """Check permissions and set the form page"""

        self.form_page = kwargs.get("form_page")

        if not get_forms_for_user(request.user).filter(pk=self.form_page.id).exists():
            raise PermissionDenied

        if self.is_export:
            data_fields = self.form_page.get_data_fields()
            # Set the export fields and the headings for spreadsheet export
            self.list_export = [field for field, label in data_fields]
            self.export_headings = dict(data_fields)

        return super().dispatch(request, *args, **kwargs)

    def get_filterset_kwargs(self):
        kwargs = super().get_filterset_kwargs()
        kwargs["queryset"] = self.get_base_queryset()
        return kwargs

    def get_base_queryset(self):
        """Return queryset of form submissions"""
        submission_class = self.form_page.get_submission_class()
        queryset = submission_class._default_manager.filter(page=self.form_page)
        return queryset

    def get_validated_ordering(self):
        """Return a dict of field names with ordering labels if ordering is valid"""
        orderable_fields = self.orderable_fields or ()
        ordering = {}
        if self.is_export:
            #  Revert to CSV order_by submit_time ascending for backwards compatibility
            default_ordering = self.ordering_csv or ()
        else:
            default_ordering = self.default_ordering or ()
        if isinstance(default_ordering, str):
            default_ordering = (default_ordering,)
        ordering_strs = self.request.GET.getlist("order_by") or list(default_ordering)
        for order in ordering_strs:
            try:
                _, prefix, field_name = order.rpartition("-")
                if field_name in orderable_fields:
                    ordering[field_name] = (
                        prefix,
                        "descending" if prefix == "-" else "ascending",
                    )
            except (IndexError, ValueError):
                continue  # invalid ordering specified, skip it
        return ordering

    def get_ordering(self):
        """Return the field or fields to use for ordering the queryset"""
        ordering = self.get_validated_ordering()
        return [values[0] + name for name, values in ordering.items()]

    def get_filename(self):
        """Returns the base filename for the generated spreadsheet data file"""
        return "{}-export-{}".format(
            self.form_page.slug, datetime.datetime.today().strftime("%Y-%m-%d")
        )

    def render_to_response(self, context, **response_kwargs):
        if self.is_export:
            return self.as_spreadsheet(
                context["submissions"], self.request.GET.get("export")
            )
        return super().render_to_response(context, **response_kwargs)

    def to_row_dict(self, item):
        """Orders the submission dictionary for spreadsheet writing"""
        row_dict = OrderedDict(
            (field, item.get_data().get(field)) for field in self.list_export
        )
        return row_dict

    def get_index_url(self):
        return reverse(self.index_url_name, args=(self.form_page.id,))

    def get_index_results_url(self):
        return reverse(self.index_results_url_name, args=(self.form_page.id,))

    def get_page_subtitle(self):
        return self.form_page.get_admin_display_title()

    def get_breadcrumbs_items(self):
        return self.breadcrumbs_items + [
            {
                "url": reverse(self.forms_index_url_name),
                "label": gettext("Forms"),
            },
            {
                "url": "",
                "label": self.get_page_title(),
                "sublabel": self.get_page_subtitle(),
            },
        ]

    def get_context_data(self, **kwargs):
        """Return context for view"""
        context = super().get_context_data(**kwargs)
        submissions = context[self.context_object_name]
        data_fields = self.form_page.get_data_fields()
        data_rows = []
        context["submissions"] = submissions
        if not self.is_export:
            # Build data_rows as list of dicts containing model_id and fields
            for submission in submissions:
                form_data = submission.get_data()
                data_row = []
                for name, label in data_fields:
                    val = form_data.get(name)
                    if isinstance(val, list):
                        val = ", ".join(val)
                    data_row.append(val)
                data_rows.append({"model_id": submission.id, "fields": data_row})
            # Build data_headings as list of dicts containing model_id and fields
            ordering_by_field = self.get_validated_ordering()
            orderable_fields = self.orderable_fields
            data_headings = []
            for name, label in data_fields:
                order_label = None
                if name in orderable_fields:
                    order = ordering_by_field.get(name)
                    if order:
                        order_label = order[1]  # 'ascending' or 'descending'
                    else:
                        order_label = "orderable"  # not ordered yet but can be
                data_headings.append(
                    {
                        "name": name,
                        "label": label,
                        "order": order_label,
                    }
                )

            context.update(
                {
                    "form_page": self.form_page,
                    "data_headings": data_headings,
                    "data_rows": data_rows,
                }
            )

        context["next_url"] = f"{self.get_index_url()}?{self.request.GET.urlencode()}"
        return context



from django.contrib.auth.models import Group
from django.test import TestCase
from django.urls import reverse

from wagtail.admin.site_summary import PagesSummaryItem
from wagtail.models import GroupPagePermission, Page, Site
from wagtail.test.testapp.models import SimplePage
from wagtail.test.utils import WagtailTestUtils


class TestPagesSummary(WagtailTestUtils, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.test_page = SimplePage(title="test", slug="test", content="test")
        cls.wagtail_root = Page.get_first_root_node()
        cls.wagtail_root.add_child(instance=cls.test_page)

        cls.test_page_group = Group.objects.create(name="Test page")

    @classmethod
    def tearDownClass(cls):
        cls.test_page.delete()
        cls.test_page_group.delete()

        super().tearDownClass()

    def setUp(self):
        self.user = self.login()
        self.request = self.client.get("/").wsgi_request

    def assertSummaryContains(self, content):
        summary = PagesSummaryItem(self.request).render_html()
        self.assertIn(content, summary)

    def assertSummaryContainsLinkToPage(self, page_pk):
        self.assertSummaryContains(reverse("wagtailadmin_explore", args=[page_pk]))

    def test_user_with_page_permissions_is_shown_panel(self):
        self.assertTrue(PagesSummaryItem(self.request).is_shown())

    def test_single_site_summary_links_to_site_root(self):
        self.assertEqual(Site.objects.count(), 1)
        site = Site.objects.first()
        self.assertSummaryContainsLinkToPage(site.root_page.pk)

    def test_multiple_sites_summary_links_to_wagtail_root(self):
        Site.objects.create(hostname="foo.com", root_page=self.wagtail_root)
        self.assertSummaryContainsLinkToPage(self.wagtail_root.pk)

    def test_no_sites_summary_links_to_wagtail_root(self):
        Site.objects.all().delete()
        self.assertSummaryContainsLinkToPage(self.wagtail_root.pk)

    def test_summary_includes_page_count_without_wagtail_root(self):
        self.assertSummaryContains(f"{Page.objects.count() - 1} Pages")

    def test_summary_shows_zero_pages_if_none_exist_except_wagtail_root(self):
        Page.objects.exclude(pk=self.wagtail_root.pk).delete()
        self.assertSummaryContains("0 Pages")

    def test_user_with_no_page_permissions_is_not_shown_panel(self):
        self.user.is_superuser = False
        self.user.save()
        self.assertFalse(PagesSummaryItem(self.request).is_shown())

    def test_user_with_limited_page_permissions(self):
        permissions = {"add", "change", "publish", "bulk_delete", "lock", "unlock"}
        self.user.is_superuser = False
        self.user.save()
        self.user.groups.add(self.test_page_group)

        for permission in permissions:
            with self.subTest(permission=permission):
                permission_obj = GroupPagePermission.objects.create(
                    group=self.test_page_group,
                    page=self.test_page,
                    permission_type=permission,
                )
                # Should link to the permission's page rather than Wagtail's root
                self.assertSummaryContainsLinkToPage(self.test_page.pk)
                # Should show the correct page count
                self.assertSummaryContains("1 Page")
                permission_obj.delete()
                self.test_page_group.page_permissions.all().delete()



from unittest import TestCase

from wagtail.images.image_operations import ImageTransform
from wagtail.images.rect import Rect, Vector


class TestTransform(TestCase):
    def test_resize(self):
        context = ImageTransform((640, 480))
        resized = context.resize((320, 240))

        vector = Vector(100, 200)
        transformed = resized.transform_vector(vector)

        self.assertEqual(transformed, Vector(50, 100))

        untransformed = resized.untransform_vector(transformed)

        self.assertEqual(untransformed, vector)

    def test_crop(self):
        context = ImageTransform((640, 480))
        cropped = context.crop(Rect(200, 100, 300, 200))

        vector = Vector(250, 150)
        transformed = cropped.transform_vector(vector)

        self.assertEqual(transformed, Vector(50, 50))

        untransformed = cropped.untransform_vector(transformed)

        self.assertEqual(untransformed, vector)

    def test_resize_then_crop(self):
        context = ImageTransform((640, 480))
        resized = context.resize((320, 240))
        cropped = resized.crop(Rect(200, 100, 300, 200))

        vector = Vector(500, 300)
        transformed = cropped.transform_vector(vector)

        self.assertEqual(transformed, Vector(50, 50))

        untransformed = cropped.untransform_vector(transformed)

        self.assertEqual(untransformed, vector)

    def test_crop_then_resize(self):
        context = ImageTransform((640, 480))
        cropped = context.crop(Rect(200, 100, 300, 200))
        resized = cropped.resize((50, 50))

        vector = Vector(250, 150)
        transformed = resized.transform_vector(vector)

        self.assertEqual(transformed, Vector(25, 25))

        untransformed = resized.untransform_vector(transformed)

        self.assertEqual(untransformed, vector)



# Generated by Django 4.0.10 on 2023-11-10 01:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("searchtests", "0002_bookunindexed"),
    ]

    operations = [
        migrations.AddField(
            model_name="book",
            name="summary",
            field=models.TextField(blank=True),
        ),
    ]



import itertools

from django import forms
from django.test import TestCase

from wagtail.contrib.forms.forms import FormBuilder
from wagtail.contrib.forms.utils import get_field_clean_name
from wagtail.models import Page
from wagtail.test.testapp.models import (
    ExtendedFormField,
    FormBuilderWithCustomWidget,
    FormField,
    FormPage,
    FormPageWithCustomFormBuilder,
)


class TestFormBuilder(TestCase):
    def setUp(self):
        # Create a form page
        home_page = Page.objects.get(url_path="/home/")

        self.form_page = home_page.add_child(
            instance=FormPage(
                title="Contact us",
                slug="contact-us",
                to_address="to@email.com",
                from_address="from@email.com",
                subject="The subject",
            )
        )

        FormField.objects.create(
            page=self.form_page,
            sort_order=1,
            label="Your name",
            field_type="singleline",
            required=True,
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=2,
            label="Your message",
            field_type="multiline",
            required=True,
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=2,
            label="Your birthday",
            field_type="date",
            required=True,
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=2,
            label="Your birthtime :)",
            field_type="datetime",
            required=True,
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=1,
            label="Your email",
            field_type="email",
            required=True,
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=2,
            label="Your homepage",
            field_type="url",
            required=True,
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=2,
            label="Your favourite number",
            field_type="number",
            required=True,
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=2,
            label="Your favourite text editors",
            field_type="multiselect",
            required=True,
            choices="vim,nano,emacs",
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=2,
            label="Your favourite Python IDEs",
            field_type="dropdown",
            required=True,
            choices="PyCharm,vim,nano",
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=2,
            label="Ὕour favourite Ρython ÏÐÈ",  # unicode example
            help_text="Choose one",
            field_type="radio",
            required=True,
            choices="PyCharm,vim,nano",
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=3,
            label="Your choices",
            field_type="checkboxes",
            required=False,
            choices="foo,bar,baz",
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=3,
            label="I agree to the Terms of Use",
            field_type="checkbox",
            required=True,
        )
        FormField.objects.create(
            page=self.form_page,
            sort_order=1,
            label="A Hidden Field",
            field_type="hidden",
            required=False,
        )

        # Create a form builder
        self.fb = FormBuilder(self.form_page.get_form_fields())

    def test_fields(self):
        """
        This tests that all fields were added to the form with the correct types
        """
        form_class = self.fb.get_form_class()

        # All fields are present in form
        field_names = form_class.base_fields.keys()
        self.assertIn("your_name", field_names)
        self.assertIn("your_message", field_names)
        self.assertIn("your_birthday", field_names)
        self.assertIn("your_birthtime", field_names)
        self.assertIn("your_email", field_names)
        self.assertIn("your_homepage", field_names)
        self.assertIn("your_favourite_number", field_names)
        self.assertIn("your_favourite_text_editors", field_names)
        self.assertIn("your_favourite_python_ides", field_names)
        self.assertIn("u03a5our_favourite_u03a1ython_ixd0e", field_names)
        self.assertIn("your_choices", field_names)
        self.assertIn("i_agree_to_the_terms_of_use", field_names)
        self.assertIn("a_hidden_field", field_names)

        # All fields have proper type
        self.assertIsInstance(form_class.base_fields["your_name"], forms.CharField)
        self.assertIsInstance(form_class.base_fields["your_message"], forms.CharField)
        self.assertIsInstance(form_class.base_fields["your_birthday"], forms.DateField)
        self.assertIsInstance(
            form_class.base_fields["your_birthtime"], forms.DateTimeField
        )
        self.assertIsInstance(form_class.base_fields["your_email"], forms.EmailField)
        self.assertIsInstance(form_class.base_fields["your_homepage"], forms.URLField)
        self.assertIsInstance(
            form_class.base_fields["your_favourite_number"], forms.DecimalField
        )
        self.assertIsInstance(
            form_class.base_fields["your_favourite_text_editors"],
            forms.MultipleChoiceField,
        )
        self.assertIsInstance(
            form_class.base_fields["your_favourite_python_ides"], forms.ChoiceField
        )
        self.assertIsInstance(
            form_class.base_fields["u03a5our_favourite_u03a1ython_ixd0e"],
            forms.ChoiceField,
        )
        self.assertIsInstance(
            form_class.base_fields["your_choices"], forms.MultipleChoiceField
        )
        self.assertIsInstance(
            form_class.base_fields["i_agree_to_the_terms_of_use"], forms.BooleanField
        )
        self.assertIsInstance(form_class.base_fields["a_hidden_field"], forms.CharField)

        # Some fields have non-default widgets
        self.assertIsInstance(
            form_class.base_fields["your_message"].widget, forms.Textarea
        )
        self.assertIsInstance(
            form_class.base_fields["u03a5our_favourite_u03a1ython_ixd0e"].widget,
            forms.RadioSelect,
        )
        self.assertIsInstance(
            form_class.base_fields["your_choices"].widget, forms.CheckboxSelectMultiple
        )
        self.assertIsInstance(
            form_class.base_fields["a_hidden_field"].widget, forms.HiddenInput
        )

    def test_unsaved_fields_in_form_builder_formfields(self):
        """Ensure unsaved FormField instances are added to FormBuilder.formfields dict
        with a clean_name as the key.
        """
        unsaved_field_1 = FormField(
            page=self.form_page,
            sort_order=14,
            label="Unsaved field 1",
            field_type="singleline",
            required=True,
        )
        self.form_page.form_fields.add(unsaved_field_1)

        unsaved_field_2 = FormField(
            page=self.form_page,
            sort_order=15,
            label="Unsaved field 2",
            field_type="singleline",
            required=True,
        )
        self.form_page.form_fields.add(unsaved_field_2)

        fb = FormBuilder(self.form_page.get_form_fields())
        self.assertIn(get_field_clean_name(unsaved_field_1.label), fb.formfields)
        self.assertIn(get_field_clean_name(unsaved_field_2.label), fb.formfields)

    def test_newline_value_separation_in_choices(self):
        """
        Ensure that choices can be separated either by newlines or by commas.
        """
        field_types = ["multiselect", "dropdown", "checkboxes", "radio"]
        testdata = [
            ("red\ngreen\nblue", ["red", "green", "blue"]),
            ("red\rgreen\rblue", ["red", "green", "blue"]),
            ("red\r\ngreen\r\nblue", ["red", "green", "blue"]),
            ("red\r\ngreen\nblue", ["red", "green", "blue"]),
            ("red\ngreen\nblue\n", ["red", "green", "blue"]),
            ("\nred\ngreen\nblue", ["red", "green", "blue"]),
            ("red,\ngreen \n blue", ["red", "green", "blue"]),
            ("red  ,\ngreen \n blue", ["red", "green", "blue"]),
            ("  red  \ngreen\nblue", ["red", "green", "blue"]),
            ("red\n,green\nblue", ["red", ",green", "blue"]),
            ("red\ngreen, blue\nyellow", ["red", "green, blue", "yellow"]),
        ]
        for field_type, (choices, expected) in itertools.product(field_types, testdata):
            field = FormField(field_type=field_type, choices=choices, label="test")
            builder = FormBuilder([field])
            form_field = builder.formfields["test"]
            with self.subTest(field_type=field_type, choices=choices):
                self.assertEqual(
                    [choice for choice, _ in form_field.choices],
                    expected,
                )

    def test_newline_value_separation_in_default_value(self):
        """
        Ensure that default values can be separated by newlines.
        """
        for choices, default_value, expected in [
            ("a\nb\nc", "a", ["a"]),
            ("a\nb\nc", "a\nc", ["a", "c"]),
            ("a\rb\rc", "a\rc", ["a", "c"]),
            ("a\r\nb\r\nc", "a\r\nc", ["a", "c"]),
            ("a\nb\nc", "a\r\nc", ["a", "c"]),
            ("a\nb\rc", "a\r\nc", ["a", "c"]),
            ("a\nb\nc", "a\nd", ["a", "d"]),
        ]:
            field = FormField(
                label="test",
                field_type="checkboxes",
                choices=choices,
                default_value=default_value,
            )
            builder = FormBuilder([field])
            form_field = builder.formfields["test"]
            with self.subTest(choices=choices, default_value=default_value):
                self.assertEqual(form_field.initial, expected)

    def test_custom_widget(self):
        """
        All builtin field types should be able to receive a custom widget
        """
        self.form_page.form_builder = FormBuilderWithCustomWidget
        form = self.form_page.get_form(auto_id=None)
        for fieldname, expected_render in [
            (
                "your_name",
                '<input type="text" name="your_name" maxlength="255" class="custom">',
            ),
            ("your_message", '<input type="text" name="your_message" class="custom">'),
            (
                "your_birthday",
                '<input type="text" name="your_birthday" class="custom">',
            ),
            (
                "your_birthtime",
                '<input type="text" name="your_birthtime" class="custom">',
            ),
            (
                "your_email",
                '<input type="text" name="your_email" maxlength="320" class="custom">',
            ),
            (
                "your_homepage",
                '<input type="text" name="your_homepage" class="custom">',
            ),
            (
                "your_favourite_number",
                '<input type="text" name="your_favourite_number" class="custom">',
            ),
            (
                "your_favourite_text_editors",
                '<input type="text" name="your_favourite_text_editors" class="custom">',
            ),
            (
                "your_favourite_python_ides",
                '<input type="text" name="your_favourite_python_ides" class="custom">',
            ),
            (
                "u03a5our_favourite_u03a1ython_ixd0e",
                '<input type="text" name="u03a5our_favourite_u03a1ython_ixd0e" class="custom">',
            ),
            (
                "your_choices",
                '<input type="text" name="your_choices" value="[\'\']" class="custom">',
            ),
            (
                "i_agree_to_the_terms_of_use",
                '<input type="text" name="i_agree_to_the_terms_of_use" class="custom">',
            ),
            (
                "a_hidden_field",
                '<input type="text" name="a_hidden_field" class="custom">',
            ),
        ]:
            with self.subTest(field=fieldname):
                form.fields[fieldname].required = False  # makes testing easier
                self.assertHTMLEqual(form[fieldname].as_widget(), expected_render)


class TestCustomFormBuilder(TestCase):
    def setUp(self):
        # Create a form page
        home_page = Page.objects.get(url_path="/home/")

        self.form_page = home_page.add_child(
            instance=FormPageWithCustomFormBuilder(
                title="IT Support Request",
                slug="it-support-request",
                to_address="it@jenkins.com",
                from_address="support@jenkins.com",
                subject="Support Request Submitted",
            )
        )
        ExtendedFormField.objects.create(
            page=self.form_page,
            sort_order=1,
            label="Name",
            field_type="singleline",
            required=True,
        )

    def test_using_custom_form_builder(self):
        """Tests that charfield max_length is 120 characters."""
        form_class = self.form_page.get_form_class()
        form = form_class()
        # check name field exists
        self.assertIsInstance(form.base_fields["name"], forms.CharField)
        # check max_length is set
        self.assertEqual(form.base_fields["name"].max_length, 120)

    def test_adding_custom_field(self):
        """Tests that we can add the ipaddress field, which is an extended choice."""
        ExtendedFormField.objects.create(
            page=self.form_page,
            sort_order=1,
            label="Device IP Address",
            field_type="ipaddress",
            required=True,
        )
        form_class = self.form_page.get_form_class()
        form = form_class()
        # check ip address field used
        self.assertIsInstance(
            form.base_fields["device_ip_address"], forms.GenericIPAddressField
        )

    def test_unsaved_fields_in_form_builder_formfields_with_clean_name_override(self):
        """
        Ensure unsaved FormField instances are added to FormBuilder.formfields dict
        with a clean_name that uses the `get_field_clean_name` method that can be overridden.
        """

        unsaved_field_1 = ExtendedFormField(
            page=self.form_page,
            sort_order=14,
            label="Unsaved field 1",
            field_type="number",
            required=True,
        )
        self.form_page.form_fields.add(unsaved_field_1)

        unsaved_field_2 = ExtendedFormField(
            page=self.form_page,
            sort_order=15,
            label="duplicate (suffix removed)",
            field_type="singleline",
            required=True,
        )
        self.form_page.form_fields.add(unsaved_field_2)

        form_class = self.form_page.get_form_class()
        form = form_class()

        # See ExtendedFormField get_field_clean_name method
        self.assertIn("number_field--unsaved_field_1", form.base_fields)
        self.assertIn("test duplicate", form.base_fields)



from django.test import TransactionTestCase

from wagtail.test.utils import WagtailTestUtils


class AdminAPITestCase(WagtailTestUtils, TransactionTestCase):
    def setUp(self):
        super().setUp()
        self.user = self.login()



from django import forms
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from wagtail.users.views.bulk_actions.user_bulk_action import UserBulkAction
from wagtail.users.views.users import change_user_perm


class ActivityForm(forms.Form):
    mark_as_active = forms.TypedChoiceField(
        choices=((True, _("Active")), (False, _("Inactive"))),
        label=_("Mark as active"),
        widget=forms.RadioSelect,
        coerce=lambda x: x == "True",
    )


class SetActiveStateBulkAction(UserBulkAction):
    display_name = _("Set active state")
    action_type = "set_active_state"
    aria_label = _("Change the active state for selected users")
    template_name = "wagtailusers/bulk_actions/confirm_bulk_set_active_state.html"
    action_priority = 20
    form_class = ActivityForm

    def check_perm(self, obj):
        return self.request.user.has_perm(change_user_perm)

    def get_execution_context(self):
        return {
            "mark_as_active": self.cleaned_form.cleaned_data["mark_as_active"],
            "user": self.request.user,
            "model": self.model,
        }

    def get_actionable_objects(self):
        objects, objects_without_access = super().get_actionable_objects()
        user = self.request.user
        users = list(filter(lambda x: x.pk != user.pk, objects))
        if len(objects) != len(users):
            objects_without_access["mark_self_as_inactive"] = [user]
        return users, objects_without_access

    @classmethod
    def execute_action(cls, objects, mark_as_active=False, model=None, **kwargs):
        if model is None:
            model = cls.get_default_model()
        user = kwargs.get("user", None)
        if user is not None:
            objects = list(filter(lambda x: x.pk != user.pk, objects))
        num_parent_objects = model.objects.filter(
            pk__in=[obj.pk for obj in objects]
        ).update(is_active=mark_as_active)
        return num_parent_objects, 0

    def get_success_message(self, num_parent_objects, num_child_objects):
        if self.cleaned_form.cleaned_data["mark_as_active"]:
            return ngettext(
                "%(num_parent_objects)d user has been marked as active",
                "%(num_parent_objects)d users have been marked as active",
                num_parent_objects,
            ) % {
                "num_parent_objects": num_parent_objects,
            }
        else:
            return ngettext(
                "%(num_parent_objects)d user has been marked as inactive",
                "%(num_parent_objects)d users have been marked as inactive",
                num_parent_objects,
            ) % {
                "num_parent_objects": num_parent_objects,
            }



from django.conf import settings
from django.db import transaction
from django.db.models.signals import post_delete, post_save

from wagtail.images import get_image_model
from wagtail.tasks import delete_file_from_storage_task

from .tasks import set_image_focal_point_task


def post_delete_file_cleanup(instance, **kwargs):
    transaction.on_commit(
        lambda: delete_file_from_storage_task.enqueue(
            instance.file.storage.deconstruct(), instance.file.name
        )
    )


def post_delete_purge_rendition_cache(instance, **kwargs):
    instance.purge_from_cache()


def post_save_image_feature_detection(instance, **kwargs):
    if getattr(settings, "WAGTAILIMAGES_FEATURE_DETECTION_ENABLED", False):
        # Make sure the image is not from a fixture
        # and the image doesn't already have a focal point
        if kwargs["raw"] is False and not instance.has_focal_point():
            # Set the focal point
            set_image_focal_point_task.enqueue(
                instance._meta.app_label, instance._meta.model_name, str(instance.pk)
            )


def register_signal_handlers():
    Image = get_image_model()
    Rendition = Image.get_rendition_model()

    post_save.connect(post_save_image_feature_detection, sender=Image)
    post_delete.connect(post_delete_file_cleanup, sender=Image)
    post_delete.connect(post_delete_file_cleanup, sender=Rendition)
    post_delete.connect(post_delete_purge_rendition_cache, sender=Rendition)



import factory
from factory.django import DjangoModelFactory

from wagtail.test.utils import wagtail_factories

from . import models


class SimpleStructBlockFactory(wagtail_factories.StructBlockFactory):
    char1 = "Char Block 1"
    char2 = "Char Block 2"

    class Meta:
        model = models.SimpleStructBlock


class SimpleStreamBlockFactory(wagtail_factories.StreamBlockFactory):
    char1 = "Char Block 1"
    char2 = "Char Block 2"

    class Meta:
        model = models.SimpleStreamBlock


class NestedStructBlockFactory(wagtail_factories.StructBlockFactory):
    char1 = "Char Block 1"
    struct1 = factory.SubFactory(SimpleStructBlockFactory)
    stream1 = factory.SubFactory(SimpleStreamBlockFactory)
    list1 = wagtail_factories.ListBlockFactory(wagtail_factories.CharBlockFactory)

    class Meta:
        model = models.NestedStructBlock


class NestedStreamBlockFactory(wagtail_factories.StreamBlockFactory):
    char1 = "Char Block 1"
    struct1 = factory.SubFactory(SimpleStructBlockFactory)
    stream1 = factory.SubFactory(SimpleStreamBlockFactory)
    list1 = wagtail_factories.ListBlockFactory(wagtail_factories.CharBlockFactory)

    class Meta:
        model = models.NestedStreamBlock


class BaseStreamBlockFactory(wagtail_factories.StreamBlockFactory):
    char1 = "Char Block 1"
    char2 = "Char Block 2"
    simplestruct = factory.SubFactory(SimpleStructBlockFactory)
    simplestream = factory.SubFactory(SimpleStreamBlockFactory)
    simplelist = wagtail_factories.ListBlockFactory(wagtail_factories.CharBlockFactory)
    nestedstruct = factory.SubFactory(NestedStructBlockFactory)
    nestedstream = factory.SubFactory(NestedStreamBlockFactory)
    nestedlist_struct = wagtail_factories.ListBlockFactory(SimpleStructBlockFactory)
    nestedlist_stream = wagtail_factories.ListBlockFactory(SimpleStreamBlockFactory)

    class Meta:
        model = models.BaseStreamBlock


class SampleModelFactory(DjangoModelFactory):
    content = wagtail_factories.StreamFieldFactory(BaseStreamBlockFactory)

    class Meta:
        model = models.SampleModel


class SamplePageFactory(wagtail_factories.PageFactory):
    content = wagtail_factories.StreamFieldFactory(BaseStreamBlockFactory)

    class Meta:
        model = models.SamplePage



# Generated by Django 2.1.13 on 2019-10-17 13:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0043_lock_fields"),
    ]

    operations = [
        migrations.AlterField(
            model_name="grouppagepermission",
            name="permission_type",
            field=models.CharField(
                choices=[
                    ("add", "Add/edit pages you own"),
                    ("edit", "Edit any page"),
                    ("publish", "Publish any page"),
                    ("bulk_delete", "Delete pages with children"),
                    ("lock", "Lock/unlock pages you've locked"),
                    ("unlock", "Unlock any page"),
                ],
                max_length=20,
                verbose_name="permission type",
            ),
        ),
    ]



from django.dispatch import Signal

# Generic object signals

# provides args: instance, revision
published = Signal()

# provides args: instance
unpublished = Signal()


# Page signals

# provides args: instance, revision
page_published = Signal()

# provides args: instance
page_unpublished = Signal()

# provides args: instance, instance_before
page_slug_changed = Signal()

# provides args: instance, parent_page_before, parent_page_after, url_path_before, url_path_after
pre_page_move = Signal()

# provides args: instance, parent_page_before, parent_page_after, url_path_before, url_path_after
post_page_move = Signal()


# Workflow signals

# provides args: instance, user
workflow_approved = Signal()

# provides args: instance, user
workflow_rejected = Signal()

# provides args: instance, user
workflow_cancelled = Signal()

# provides args: instance, user
workflow_submitted = Signal()


# Workflow task signals

# provides args: instance, user
task_approved = Signal()

# provides args: instance, user
task_rejected = Signal()

# provides args: instance, user
task_submitted = Signal()

# provides args: instance, user
task_cancelled = Signal()


# Locale signals

# Like pre_delete, but sent on deletion before on_delete validation is applied.
# Currently only sent by the Locale model.
# Required as a workaround for https://code.djangoproject.com/ticket/6870
# provides args: sender, instance
pre_validate_delete = Signal()

# Translation signals
# provides args: sender, source_obj, target_obj
copy_for_translation_done = Signal()

# Admin signals
# provides args: page, parent
init_new_page = Signal()



from django.forms import widgets


class AdminAutoHeightTextInput(widgets.Textarea):
    def __init__(self, attrs=None):
        # Use more appropriate rows default, given autosize will alter this anyway
        default_attrs = {
            "rows": 1,
            "data-controller": "w-autosize",
        }
        if attrs:
            default_attrs.update(attrs)

        # add a w-field__autosize classname
        try:
            default_attrs["class"] += " w-field__autosize"
        except KeyError:
            default_attrs["class"] = "w-field__autosize"

        super().__init__(default_attrs)



from django.conf import settings
from django.urls import reverse
from django.utils.functional import cached_property

from wagtail.admin import messages

# Retain backwards compatibility for imports
from wagtail.admin.utils import (  # noqa: F401
    get_latest_str,
    get_valid_next_url_from_request,
)
from wagtail.permissions import page_permission_policy


def get_breadcrumbs_items_for_page(
    page,
    user,
    url_name="wagtailadmin_explore",
    root_url_name="wagtailadmin_explore_root",
    include_self=True,
    querystring_value="",
):
    # find the closest common ancestor of the pages that this user has direct explore permission
    # (i.e. add/edit/publish/lock) over; this will be the root of the breadcrumb
    cca = page_permission_policy.explorable_root_instance(user)
    if not cca:
        return []

    pages = (
        page.get_ancestors(inclusive=include_self)
        .descendant_of(cca, inclusive=True)
        .specific(defer=True)
    )

    items = []
    for page in pages:
        if page.is_root() and root_url_name:
            url = reverse(root_url_name)
        else:
            url = reverse(url_name, args=(page.id,))
        items.append({"url": url + querystring_value, "label": get_latest_str(page)})

    return items


class GenericPageBreadcrumbsMixin:
    """
    A mixin that allows a view for pages that extends a generic view to combine
    the page explorer breadcrumbs with the generic view's breadcrumbs.

    This is done by generating the explorer breadcrumbs items for the page as a
    normalised breadcrumbs items list, and then concatenating that with the last
    item of the generic view's generated breadcrumbs items.
    """

    breadcrumbs_items_to_take = 1

    @cached_property
    def breadcrumbs_items(self):
        return get_breadcrumbs_items_for_page(self.object, self.request.user)

    def get_breadcrumbs_items(self):
        # The generic view tends to generate breadcrumbs with items such as
        # IndexView > EditView > CurrentView,
        # but we don't want that because we want the preceding items to be links
        # to the explore view of the page's ancestors for consistency with how
        # page breadcrumbs have always worked. So we only take the last N items,
        # which in most cases is the final item that links to the current view.
        # However, this can be customised in the case of generic views that are
        # nested inside another generic view.
        return (
            self.breadcrumbs_items
            + super().get_breadcrumbs_items()[-self.breadcrumbs_items_to_take :]
        )


def type_to_delete_confirmation(request, context=None):
    """
    Checks if a user has correctly typed the site name to confirm page deletion
    when using a confirmation dialog.
    """
    wagtail_site_name = getattr(settings, "WAGTAIL_SITE_NAME", "wagtail")
    if (
        request.POST.get("confirm_site_name")
        and request.POST.get("confirm_site_name") != wagtail_site_name
    ):
        messages.error(request, f"Please type '{wagtail_site_name}' to confirm.")
        return False

    return True
